
import warnings

from nanohttp import Controller, context, json, RestController, action

from restfulpy.orm import DBSession


warnings.filterwarnings('ignore', message='Unknown REQUEST_METHOD')


class RootController(Controller):

    def __call__(self, *remaining_paths):

        if context.method == 'options':
            context.response_encoding = 'utf-8'
            context.response_headers.add_header("Cache-Control", "no-cache,no-store")
            return b''

        return super().__call__(*remaining_paths)


class ModelRestController(RestController):
    __model__ = None

    @json
    def metadata(self):
        return self.__model__.json_metadata()


class JsonPatchControllerMixin:

    @action(content_type='application/json')
    def patch(self: Controller):
        """
        Set context.method
        Set context.form
        :return:
        """
        # Preserve patches
        patches = context.form
        results = []
        context.jsonpatch = True

        try:
            for patch in patches:
                context.form = patch.get('value', {})
                context.method = patch['op']

                remaining_paths = patch['path'].split('/')
                if remaining_paths and not remaining_paths[0]:
                    return_data = self()
                else:
                    return_data = self(*remaining_paths)

                results.append(return_data)

                DBSession.flush()
            DBSession.commit()
            return '[%s]' % ',\n'.join(results)
        except:
            if DBSession.is_active:
                DBSession.rollback()
            raise
        finally:
            del context.jsonpatch




import unittest
import time

from nanohttp import json, Controller, context, settings

from restfulpy.authentication import StatefulAuthenticator
from restfulpy.authorization import authorize
from restfulpy.principal import JwtPrincipal, JwtRefreshToken
from restfulpy.testing import WebAppTestCase, As
from restfulpy.tests.helpers import MockupApplication


class MockupMember:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


roles = ['admin', 'test']


class MockupStatefulAuthenticator(StatefulAuthenticator):
    def validate_credentials(self, credentials):
        email, password = credentials
        if password == 'test':
            return MockupMember(id=1, email=email, roles=['admin', 'test'])

    def create_refresh_principal(self, member_id=None):
        return JwtRefreshToken(dict(
            id=member_id
        ))

    def create_principal(self, member_id=None, session_id=None):
        return JwtPrincipal(dict(id=1, email='test@example.com', roles=roles, sessionId='1'))


class Root(Controller):

    @json
    def login(self):
        principal = context.application.__authenticator__.login((context.form['email'], context.form['password']))
        return dict(token=principal.dump())

    @json
    @authorize
    def me(self):
        return context.identity.payload

    @json
    @authorize
    def invalidate_token(self):
        context.application.__authenticator__.invalidate_member(1)
        return context.identity.payload

    @json(verbs='delete')
    @authorize
    def logout(self):
        context.application.__authenticator__.logout()
        return {}


class StatefulAuthenticatorTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', Root(), authenticator=MockupStatefulAuthenticator())

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge("""
            jwt:
              max_age: .3
              refresh_token:
                max_age: 3
        """)

    def test_invalidate_token(self):
        response, headers = self.request('ALL', 'GET', '/login', json=dict(email='test@example.com', password='test'))
        refresh_token = headers['Set-Cookie'].split('; ')[0]
        self.assertIn('token', response)
        self.assertTrue(refresh_token.startswith('refresh-token='))

        # Login on client
        token = response['token']
        self.wsgi_app.jwt_token = token

        # Request a protected resource to ensure authenticator is working well
        response, ___ = self.request(As.member, 'GET', '/me', headers={'Cookie': refresh_token})
        self.assertListEqual(response['roles'], roles)

        # Invalidating the token by server
        roles.append('god')
        response, headers = self.request(As.member, 'GET', '/invalidate_token', headers={'Cookie': refresh_token})
        self.assertListEqual(response['roles'], roles)
        self.assertIn('X-New-JWT-Token', headers)

        # Invalidating the token by server after the token has been expired expired, with appropriate cookies.
        time.sleep(1)
        response, headers = self.request(
            As.member, 'GET', '/invalidate_token',
            headers={
                'Cookie': refresh_token
            }
        )
        self.assertIn('X-New-JWT-Token', headers)
        self.assertIsNotNone(headers['X-New-JWT-Token'])
        self.wsgi_app.jwt_token = headers['X-New-JWT-Token']
        self.request(As.member, 'GET', '/me', headers={'Cookie': refresh_token})

    def test_logout(self):
        response, headers = self.request('ALL', 'POST', '/login', json=dict(email='test@example.com', password='test'))
        self.assertIn('token', response)
        self.assertEqual(headers['X-Identity'], '1')
        self.wsgi_app.jwt_token = response['token']
        response, headers = self.request('ALL', 'DELETE', '/logout')
        self.assertEqual(headers['X-Identity'], '')


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

from sqlalchemy import Integer, Unicode
from nanohttp import settings

from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication
from restfulpy.orm import DeclarativeBase, Field, DBSession, ApproveRequiredMixin


class ApproveRequiredObject(ApproveRequiredMixin, DeclarativeBase):
    __tablename__ = 'approve_required_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))


class ApproveRequiredMixinTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', None)
    __configuration__ = '''
    db:
      uri: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)

    def test_approve_required_mixin(self):
        # noinspection PyArgumentList
        object1 = ApproveRequiredObject(
            title='object 1',
        )

        DBSession.add(object1)
        DBSession.commit()
        self.assertFalse(object1.is_approved)
        self.assertEqual(DBSession.query(ApproveRequiredObject).filter(ApproveRequiredObject.is_approved).count(), 0)

        object1.is_approved = True
        self.assertTrue(object1.is_approved)
        DBSession.commit()
        object1 = DBSession.query(ApproveRequiredObject).one()
        self.assertTrue(object1.is_approved)

        json = object1.to_dict()
        self.assertIn('isApproved', json)

        self.assertEqual(DBSession.query(ApproveRequiredObject).filter(ApproveRequiredObject.is_approved).count(), 1)
        self.assertEqual(ApproveRequiredObject.filter_approved().count(), 1)

        self.assertFalse(ApproveRequiredObject.import_value(ApproveRequiredObject.is_approved, 'false'))
        self.assertFalse(ApproveRequiredObject.import_value(ApproveRequiredObject.is_approved, 'FALSE'))
        self.assertFalse(ApproveRequiredObject.import_value(ApproveRequiredObject.is_approved, 'False'))
        self.assertTrue(ApproveRequiredObject.import_value(ApproveRequiredObject.is_approved, 'true'))
        self.assertTrue(ApproveRequiredObject.import_value(ApproveRequiredObject.is_approved, 'TRUE'))
        self.assertTrue(ApproveRequiredObject.import_value(ApproveRequiredObject.is_approved, 'True'))

        self.assertEqual(ApproveRequiredObject.import_value(ApproveRequiredObject.title, 'title'), 'title')


if __name__ == '__main__':  # pragma: no cover
    unittest.main()





DOC_HEADER = """

| Parameter | Value       |
| --------- | ----------- |
| Author    | TestEngine  |
| Version   | %(version)s |
| Status    | Implemented |

"""


DOC_LEGEND = """

#### Legend


##### Paging

| Param  | Meaning            |
| ------ | ------------------ |
| take   | Rows per page      |
| skip   | Skip N rows        |


##### Search & Filtering

You can search and filter the result via query-string:

        /path/to/resource?field=[op]value1[,value2]

| Operator  | Meaning | Example    |
| --------- | ------- | -----------|
|           | =       | id=2       |
| !         | !=      | id=!2      |
| >         | >       | id=>2      |
| >=        | >=      | id=>=2     |
| <         | <       | id=<2      |
| <=        | <=      | id=<=2     |
| %         | LIKE    | title=%us  |
| %~        | ILIKE   | title=%~us |
| ^         | IN      | id=^2,3,4  |
| !^        | NOT IN  | id=!^2,3,4 |
| ~         | BETWEEN | id=~2,9    |


##### Sorting

You can sort like this:


        /path/to/resource?sort=[op]value


| Operator  | Meaning |
| --------- | ------- |
|           | ASC     |
| \-        | DESC    |

"""


class As:
    anonymouse = 'Anonymouse'
    member = 'Member'
    admin = 'Admin'




import io
import unittest
from os.path import dirname, abspath, join

from nanohttp import settings, configure

from restfulpy.messaging.providers import SmtpProvider
from restfulpy.testing.mockup import smtp_server


HERE = abspath(dirname(__file__))


class SmtpProviderTestCase(unittest.TestCase):
    __configuration__ = '''
        smtp:
          host: smtp.example.com
          port: 587
          username: user@example.com
          password: password
          local_hostname: localhost
          tls: false
          auth: false
          ssl: false
        messaging:
            mako_modules_directory: %s
            template_dirs: 
              - %s
        ''' % (
        join(HERE, 'data', 'mako_modules'),
        join(HERE, 'templates'),
    )

    @classmethod
    def setUpClass(cls):
        configure(init_value=cls.__configuration__, force=True)

    def test_smtp_provider(self):
        with smtp_server() as (server, bind):
            settings.smtp.host = bind[0]
            settings.smtp.port = bind[1]

            # Without templates
            SmtpProvider().send(
                'test@example.com',
                'test@example.com',
                'Simple test body',
                cc='test@example.com',
                bcc='test@example.com'
            )

            # With template
            SmtpProvider().send(
                'test@example.com',
                'test@example.com',
                {},
                template_filename='test-email-template.mako'
            )

            # With attachments
            attachment = io.BytesIO(b'This is test attachment file')
            attachment.name = 'path/to/file.txt'
            SmtpProvider().send(
                'test@example.com',
                'test@example.com',
                'email body with Attachment',
                attachments=[attachment]
            )


if __name__ == '__main__':  # pragma: no cover
    unittest.main()






class RestfulException(Exception):
    def to_json(self):
        return {
            'type': str(type(self)),
            'content': str(self),
        }


class UnsupportedError(Exception):
    """Used to indicate we're not supporting a feature and or method.

    .. versionadded:: 0.11.0

    """
    pass



import unittest

from nanohttp import settings

from restfulpy.orm import DBSession, Field
from restfulpy.messaging import BaseEmail, Email, create_messenger
from restfulpy.testing.helpers import FakeJson
from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication


BaseEmail.metadata.remove(Email.__table__)


class Welcome(BaseEmail):
    __tablename__ = 'welcome'

    __mapper_args__ = {
        'polymorphic_identity': 'welcome'
    }

    body = Field(FakeJson, json='body')

    @property
    def email_body(self):
        return self.body

    @property
    def template_filename(self):
        return None


class MessagingModelTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', None)
    __configuration__ = '''
    db:
      uri: sqlite://    # In memory DB
      echo: false
      
    messaging:
      default_sender: test@example.com
      default_messenger: restfulpy.testing.helpers.MockupMessenger
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)

    def test_messaging_model(self):
        mockup_messenger = create_messenger()

        # noinspection PyArgumentList
        message = Welcome(
            to='test@example.com',
            subject='Test Subject',
            body={'msg': 'Hello'}
        )

        DBSession.add(message)
        DBSession.commit()

        message.do_({})

        # noinspection PyUnresolvedReferences
        self.assertDictEqual(mockup_messenger.last_message, {
            'body': {'msg': 'Hello'},
            'subject': 'Test Subject',
            'to': 'test@example.com'
        })


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




import itsdangerous
import redis
from nanohttp import context, HttpBadRequest, settings

from restfulpy.principal import JwtPrincipal, JwtRefreshToken


class Authenticator:
    """
    An extendable stateless abstract class for encapsulating all stuff about authentication
    """

    token_key = 'HTTP_AUTHORIZATION'
    refresh_token_key = 'refresh-token'
    token_response_header = 'X-New-JWT-Token'
    identity_response_header = 'X-Identity'

    def create_principal(self, member_id=None, session_id=None):  # pragma: no cover
        raise NotImplementedError()

    def create_refresh_principal(self, member_id=None):  # pragma: no cover
        raise NotImplementedError()

    def validate_credentials(self, credentials):  # pragma: no cover
        raise NotImplementedError()

    def setup_response_headers(self, new_principal):
        if self.token_response_header in context.response_headers:
            del context.response_headers[self.token_response_header]
        context.response_headers.add_header(self.token_response_header, new_principal.dump().decode())

    def try_refresh_token(self, session_id):
        morsel = context.cookies.get(self.refresh_token_key)
        if not morsel or morsel.value is None or not morsel.value.strip():
            self.bad()
            return

        refresh_token_encoded = morsel.value
        # Decoding the refresh token
        try:
            refresh_principal = JwtRefreshToken.load(refresh_token_encoded)
            self.ok(
                self.create_principal(member_id=refresh_principal.id, session_id=session_id),
                setup_header=True
            )
        except itsdangerous.SignatureExpired:
            self.bad()
        except itsdangerous.BadData:
            self.bad()
            raise HttpBadRequest()

    def setup_identity_response_header(self, principal):
        context.response_headers.add_header(self.identity_response_header, str(principal.id) if principal else '')

    def bad(self):
        context.identity = None
        self.setup_identity_response_header(None)

    def ok(self, principal, setup_header=False):
        context.identity = principal
        self.setup_identity_response_header(principal)
        if setup_header:
            self.setup_response_headers(principal)

    def verify_token(self, encoded_token):
        return JwtPrincipal.load(encoded_token)

    def authenticate_request(self):
        if self.token_key not in context.environ:
            self.bad()
            return

        encoded_token = context.environ[self.token_key]
        if encoded_token is None or not encoded_token.strip():
            self.bad()
            return

        try:
            self.ok(self.verify_token(encoded_token))

        except itsdangerous.SignatureExpired as ex:
            # The token has expired. So we're trying to restore it using refresh-token.
            session_id = ex.payload.get('sessionId')
            if session_id:
                self.try_refresh_token(session_id)
        except itsdangerous.BadData:
            # The token is Malformed
            self.bad()
            raise HttpBadRequest()

    def login(self, credentials):
        member = self.validate_credentials(credentials)
        if member is None:
            return None

        principal = self.create_principal(member.id)

        self.ok(principal)

        context.cookies[self.refresh_token_key] = self.create_refresh_principal(member.id).dump().decode()
        context.cookies[self.refresh_token_key]['max-age'] = settings.jwt.refresh_token.max_age
        context.cookies[self.refresh_token_key]['httponly'] = False
        context.cookies[self.refresh_token_key]['secure'] = True
        return principal

    def logout(self):
        self.bad()


# noinspection PyAbstractClass
class StatefulAuthenticator(Authenticator):
    """
    
    Redis data-model:
    
    sessions: HashMap { session_id: member_id }
    member_id: Set { session_id } 
    
    
    """

    _redis = None
    sessions_key = 'auth:sessions'
    members_key = 'auth:member:%s'

    def __init__(self):
        super().__init__()

    @staticmethod
    def create_blocking_redis_client():
        # FIXME: use unix socket
        return redis.StrictRedis(
            host=settings.authentication.redis.host,
            port=settings.authentication.redis.port,
            db=settings.authentication.redis.db,
            password=settings.authentication.redis.password
        )

    @property
    def redis(self):
        if self.__class__._redis is None:
            self.__class__._redis = self.create_blocking_redis_client()
        return self.__class__._redis

    @classmethod
    def get_member_sessions_key(cls, member_id):
        return 'auth:member:%s' % member_id

    def verify_token(self, encoded_token):
        principal = super().verify_token(encoded_token)
        if not self.validate_session(principal.session_id):
            raise itsdangerous.SignatureExpired('The token has already invalidated', principal.payload)
        return principal

    def login(self, credentials):
        principal = super().login(credentials)
        if principal is not None:
            self.register_session(principal.id, principal.session_id)
        return principal

    def logout(self):
        self.unregister_session(context.identity.session_id)
        super().logout()

    def register_session(self, member_id, session_id):
        self.redis.hset(self.sessions_key, session_id, member_id)
        self.redis.sadd(self.get_member_sessions_key(member_id), session_id)

    def unregister_session(self, session_id=None):
        session_id = session_id or context.identity.session_id
        member_id = self.redis.hget(self.sessions_key, session_id)
        self.redis.srem(self.get_member_sessions_key(member_id), session_id)
        self.redis.hdel(self.sessions_key, session_id)

    def invalidate_member(self, member_id=None):
        # store current session id if available
        current_session_id = None if context.identity is None else context.identity.session_id
        while True:
            session_id = self.redis.spop(self.get_member_sessions_key(member_id))
            if not session_id:
                break
            self.redis.hdel(self.sessions_key, session_id)
        self.redis.delete(self.get_member_sessions_key(member_id))
        if current_session_id:
            self.try_refresh_token(current_session_id)

    def validate_session(self, session_id):
        return self.redis.hexists(self.sessions_key, session_id)



import copy
import unittest

from nanohttp import context, json, settings

from restfulpy.principal import DummyIdentity
from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication
from restfulpy.validation import validate_form
from restfulpy.controllers import RestController, RootController


class ValidationTypesController(RestController):

    @json
    @validate_form(
        types={
            'typedParam1': float,
            'typedParam2': float,
            'typedParam3': float,
        },
        client={
           'types': {
               'typedParam1': int,
               'typedParam2': int
           }
        },
        admin={
           'types': {
               'typedParam1': complex,
               'typedParam4': complex
           }
        }
    )
    def post(self):
        result = copy.deepcopy(context.form)
        result.update(context.query_string)
        return result


class Root(RootController):
    validation = ValidationTypesController()


class ValidationTypesTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', Root())

    @classmethod
    def configure_app(cls):
        super().configure_app()
        settings.merge("""
            logging:
              loggers:
                default:
                  level: info
            """)

    def test_validation_types(self):
        # Test `type`
        # role -> All
        self.wsgi_app.jwt_token = DummyIdentity().dump().decode()
        result, ___ = self.request(
            'All', 'POST', '/validation',
            doc=False,
            params={
                'typedParam1': '1',
                'typedParam2': '2',
                'typedParam3': '3',
                'typedParam4': '4'
            }
        )
        self.assertEqual(type(result['typedParam1']), float)
        self.assertEqual(type(result['typedParam2']), float)
        self.assertEqual(type(result['typedParam3']), float)
        self.assertEqual(type(result['typedParam4']), str)

        self.request(
            'All', 'POST', '/validation',
            doc=False,
            params={'typedParam1': 'not_convertible'},
            expected_status=400
        )

        # -----------------------------
        # role -> Client
        self.wsgi_app.jwt_token = DummyIdentity('client').dump().decode()
        result, ___ = self.request(
            'Client', 'POST', '/validation',
            doc=False,
            params={
                'typedParam1': '1',
                'typedParam2': '2',
                'typedParam3': '3',
                'typedParam4': '4'
            }
        )
        self.assertEqual(type(result['typedParam1']), int)
        self.assertEqual(type(result['typedParam2']), int)
        self.assertEqual(type(result['typedParam3']), float)
        self.assertEqual(type(result['typedParam4']), str)

        self.request(
            'Client', 'POST', '/validation',
            doc=False,
            params={'typedParam1': 'not_convertible'},
            expected_status=400
        )

        # -----------------------------
        # role -> Admin
        self.wsgi_app.jwt_token = DummyIdentity('admin').dump().decode()
        result, ___ = self.request(
            'Admin', 'POST', '/validation',
            doc=False,
            params={
                'typedParam1': '1',
                'typedParam2': '2',
                'typedParam3': '3',
                'typedParam4': '4'
            }
        )
        # type complex is dict
        self.assertEqual(type(result['typedParam1']), dict)
        self.assertEqual(type(result['typedParam2']), float)
        self.assertEqual(type(result['typedParam3']), float)
        self.assertEqual(type(result['typedParam4']), dict)

        self.request(
            'Admin', 'POST', '/validation',
            doc=False,
            params={'typedParam1': 'not_convertible'},
            expected_status=400
        )


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import ujson

from sqlalchemy import Unicode, TypeDecorator

from restfulpy.messaging import Messenger


class MockupMessenger(Messenger):
    _last_message = None

    @property
    def last_message(self):
        return self.__class__._last_message

    @last_message.setter
    def last_message(self, value):
        self.__class__._last_message = value

    def send(
            self,
            to, subject, body,
            cc=None,
            bcc=None,
            template_string=None,
            template_filename=None,
            from_=None,
            attachments=None
    ):
        self.last_message = {
            'to': to,
            'body': body,
            'subject': subject
        }


# noinspection PyAbstractClass
class FakeJson(TypeDecorator):  # pragma: no cover
    impl = Unicode

    def process_bind_param(self, value, engine):
        return ujson.dumps(value)

    def process_result_value(self, value, engine):
        if value is None:
            return None

        return ujson.loads(value)



"""

mimetypes\_ Module
------------------

Due the python bugs:

- https://bugs.python.org/issue4963
- https://bugs.python.org/issue1043134
- https://bugs.python.org/issue6626#msg91205

The database was hardcoded here.

.. versionadded:: 0.3

"""

from os.path import splitext


def guess_extension(mimetype: str) -> str:
    return _db.get(mimetype)[0]


def guess_type(url: str):
    return _db_reverse.get(splitext(url)[1])


_db = {
    'application/andrew-inset': ('.ez',),
    'application/annodex': ('.anx',),
    'application/atom+xml': ('.atom',),
    'application/atomcat+xml': ('.atomcat',),
    'application/atomserv+xml': ('.atomsrv',),
    'application/bbolin': ('.lin',),
    'application/cu-seeme': ('.cu',),
    'application/davmount+xml': ('.davmount',),
    'application/dicom': ('.dcm',),
    'application/dsptype': ('.tsp',),
    'application/ecmascript': ('.es',),
    'application/font-sfnt': ('.otf', '.ttf',),
    'application/font-tdpfr': ('.pfr',),
    'application/font-woff': ('.woff',),
    'application/futuresplash': ('.spl',),
    'application/gzip': ('.gz',),
    'application/hta': ('.hta',),
    'application/java-archive': ('.jar',),
    'application/java-serialized-object': ('.ser',),
    'application/java-vm': ('.class',),
    'application/javascript': ('.js',),
    'application/json': ('.json',),
    'application/m3g': ('.m3g',),
    'application/mac-binhex40': ('.hqx',),
    'application/mac-compactpro': ('.cpt',),
    'application/mathematica': ('.nb', '.nbp',),
    'application/mbox': ('.mbox',),
    'application/msaccess': ('.mdb',),
    'application/msword': ('.doc', '.dot',),
    'application/mxf': ('.mxf',),
    'application/octet-stream': ('.bin', '.deploy', '.msu', '.msp',),
    'application/oda': ('.oda',),
    'application/oebps-package+xml': ('.opf',),
    'application/ogg': ('.ogx',),
    'application/onenote': ('.one', '.onetoc2', '.onetmp', '.onepkg',),
    'application/pdf': ('.pdf',),
    'application/pgp-encrypted': ('.pgp',),
    'application/pgp-keys': ('.key',),
    'application/pgp-signature': ('.sig',),
    'application/pics-rules': ('.prf',),
    'application/postscript': ('.ps', '.ai', '.eps', '.epsi', '.epsf', '.eps2', '.eps3',),
    'application/rar': ('.rar',),
    'application/rdf+xml': ('.rdf',),
    'application/rtf': ('.rtf',),
    'application/sla': ('.stl',),
    'application/smil+xml': ('.smi', '.smil',),
    'application/xhtml+xml': ('.xhtml', '.xht',),
    'application/xml': ('.xml', '.xsd',),
    'application/xslt+xml': ('.xsl', '.xslt',),
    'application/xspf+xml': ('.xspf',),
    'application/zip': ('.zip',),
    'application/vnd.android.package-archive': ('.apk',),
    'application/vnd.cinderella': ('.cdy',),
    'application/vnd.debian.binary-package': ('.deb', '.ddeb', '.udeb',),
    'application/vnd.font-fontforge-sfd': ('.sfd',),
    'application/vnd.google-earth.kml+xml': ('.kml',),
    'application/vnd.google-earth.kmz': ('.kmz',),
    'application/vnd.mozilla.xul+xml': ('.xul',),
    'application/vnd.ms-excel': ('.xls', '.xlb', '.xlt',),
    'application/vnd.ms-excel.addin.macroEnabled.12': ('.xlam',),
    'application/vnd.ms-excel.sheet.binary.macroEnabled.12': ('.xlsb',),
    'application/vnd.ms-excel.sheet.macroEnabled.12': ('.xlsm',),
    'application/vnd.ms-excel.template.macroEnabled.12': ('.xltm',),
    'application/vnd.ms-fontobject': ('.eot',),
    'application/vnd.ms-officetheme': ('.thmx',),
    'application/vnd.ms-pki.seccat': ('.cat',),
    'application/vnd.ms-powerpoint': ('.ppt', '.pps',),
    'application/vnd.ms-powerpoint.addin.macroEnabled.12': ('.ppam',),
    'application/vnd.ms-powerpoint.presentation.macroEnabled.12': ('.pptm',),
    'application/vnd.ms-powerpoint.slide.macroEnabled.12': ('.sldm',),
    'application/vnd.ms-powerpoint.slideshow.macroEnabled.12': ('.ppsm',),
    'application/vnd.ms-powerpoint.template.macroEnabled.12': ('.potm',),
    'application/vnd.ms-word.document.macroEnabled.12': ('.docm',),
    'application/vnd.ms-word.template.macroEnabled.12': ('.dotm',),
    'application/vnd.oasis.opendocument.chart': ('.odc',),
    'application/vnd.oasis.opendocument.database': ('.odb',),
    'application/vnd.oasis.opendocument.formula': ('.odf',),
    'application/vnd.oasis.opendocument.graphics': ('.odg',),
    'application/vnd.oasis.opendocument.graphics-template': ('.otg',),
    'application/vnd.oasis.opendocument.image': ('.odi',),
    'application/vnd.oasis.opendocument.presentation': ('.odp',),
    'application/vnd.oasis.opendocument.presentation-template': ('.otp',),
    'application/vnd.oasis.opendocument.spreadsheet': ('.ods',),
    'application/vnd.oasis.opendocument.spreadsheet-template': ('.ots',),
    'application/vnd.oasis.opendocument.text': ('.odt',),
    'application/vnd.oasis.opendocument.text-master': ('.odm',),
    'application/vnd.oasis.opendocument.text-template': ('.ott',),
    'application/vnd.oasis.opendocument.text-web': ('.oth',),
    'application/vnd.openxmlformats-officedocument.presentationml.presentation': ('.pptx',),
    'application/vnd.openxmlformats-officedocument.presentationml.slide': ('.sldx',),
    'application/vnd.openxmlformats-officedocument.presentationml.slideshow': ('.ppsx',),
    'application/vnd.openxmlformats-officedocument.presentationml.template': ('.potx',),
    'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet': ('.xlsx',),
    'application/vnd.openxmlformats-officedocument.spreadsheetml.template': ('.xltx',),
    'application/vnd.openxmlformats-officedocument.wordprocessingml.document': ('.docx',),
    'application/vnd.openxmlformats-officedocument.wordprocessingml.template': ('.dotx',),
    'application/vnd.rim.cod': ('.cod',),
    'application/vnd.smaf': ('.mmf',),
    'application/vnd.stardivision.calc': ('.sdc',),
    'application/vnd.stardivision.chart': ('.sds',),
    'application/vnd.stardivision.draw': ('.sda',),
    'application/vnd.stardivision.impress': ('.sdd',),
    'application/vnd.stardivision.math': ('.sdf',),
    'application/vnd.stardivision.writer': ('.sdw',),
    'application/vnd.stardivision.writer-global': ('.sgl',),
    'application/vnd.sun.xml.calc': ('.sxc',),
    'application/vnd.sun.xml.calc.template': ('.stc',),
    'application/vnd.sun.xml.draw': ('.sxd',),
    'application/vnd.sun.xml.draw.template': ('.std',),
    'application/vnd.sun.xml.impress': ('.sxi',),
    'application/vnd.sun.xml.impress.template': ('.sti',),
    'application/vnd.sun.xml.math': ('.sxm',),
    'application/vnd.sun.xml.writer': ('.sxw',),
    'application/vnd.sun.xml.writer.global': ('.sxg',),
    'application/vnd.sun.xml.writer.template': ('.stw',),
    'application/vnd.symbian.install': ('.sis',),
    'application/vnd.tcpdump.pcap': ('.cap', '.pcap',),
    'application/vnd.visio': ('.vsd', '.vst', '.vsw', '.vss',),
    'application/vnd.wap.wbxml': ('.wbxml',),
    'application/vnd.wap.wmlc': ('.wmlc',),
    'application/vnd.wap.wmlscriptc': ('.wmlsc',),
    'application/vnd.wordperfect': ('.wpd',),
    'application/vnd.wordperfect5.1': ('.wp5',),
    'application/x-123': ('.wk',),
    'application/x-7z-compressed': ('.7z',),
    'application/x-abiword': ('.abw',),
    'application/x-apple-diskimage': ('.dmg',),
    'application/x-bcpio': ('.bcpio',),
    'application/x-bittorrent': ('.torrent',),
    'application/x-cab': ('.cab',),
    'application/x-cbr': ('.cbr',),
    'application/x-cbz': ('.cbz',),
    'application/x-cdf': ('.cdf', '.cda',),
    'application/x-cdlink': ('.vcd',),
    'application/x-chess-pgn': ('.pgn',),
    'application/x-comsol': ('.mph',),
    'application/x-cpio': ('.cpio',),
    'application/x-csh': ('.csh',),
    'application/x-debian-package': ('.deb', '.udeb',),
    'application/x-director': ('.dcr', '.dir', '.dxr',),
    'application/x-dms': ('.dms',),
    'application/x-doom': ('.wad',),
    'application/x-dvi': ('.dvi',),
    'application/x-font': ('.pfa', '.pfb', '.gsf',),
    'application/x-font-pcf': ('.pcf', '.pcf.Z',),
    'application/x-freemind': ('.mm',),
    'application/x-futuresplash': ('.spl',),
    'application/x-ganttproject': ('.gan',),
    'application/x-gnumeric': ('.gnumeric',),
    'application/x-go-sgf': ('.sgf',),
    'application/x-graphing-calculator': ('.gcf',),
    'application/x-gtar': ('.gtar',),
    'application/x-gtar-compressed': ('.tgz', '.taz',),
    'application/x-hdf': ('.hdf',),
    'application/x-hwp': ('.hwp',),
    'application/x-ica': ('.ica',),
    'application/x-info': ('.info',),
    'application/x-internet-signup': ('.ins', '.isp',),
    'application/x-iphone': ('.iii',),
    'application/x-iso9660-image': ('.iso',),
    'application/x-jam': ('.jam',),
    'application/x-java-jnlp-file': ('.jnlp',),
    'application/x-jmol': ('.jmz',),
    'application/x-kchart': ('.chrt',),
    'application/x-killustrator': ('.kil',),
    'application/x-koan': ('.skp', '.skd', '.skt', '.skm',),
    'application/x-kpresenter': ('.kpr', '.kpt',),
    'application/x-kspread': ('.ksp',),
    'application/x-kword': ('.kwd', '.kwt',),
    'application/x-latex': ('.latex',),
    'application/x-lha': ('.lha',),
    'application/x-lyx': ('.lyx',),
    'application/x-lzh': ('.lzh',),
    'application/x-lzx': ('.lzx',),
    'application/x-maker': ('.frm', '.maker', '.frame', '.fm', '.fb', '.book', '.fbdoc',),
    'application/x-mif': ('.mif',),
    'application/x-mpegURL': ('.m3u8',),
    'application/x-ms-application': ('.application',),
    'application/x-ms-manifest': ('.manifest',),
    'application/x-ms-wmd': ('.wmd',),
    'application/x-ms-wmz': ('.wmz',),
    'application/x-msdos-program': ('.com', '.exe', '.bat', '.dll',),
    'application/x-msi': ('.msi',),
    'application/x-netcdf': ('.nc',),
    'application/x-ns-proxy-autoconfig': ('.pac',),
    'application/x-nwc': ('.nwc',),
    'application/x-object': ('.o',),
    'application/x-oz-application': ('.oza',),
    'application/x-pkcs7-certreqresp': ('.p7r',),
    'application/x-pkcs7-crl': ('.crl',),
    'application/x-python-code': ('.pyc', '.pyo',),
    'application/x-qgis': ('.qgs', '.shp', '.shx',),
    'application/x-quicktimeplayer': ('.qtl',),
    'application/x-rdp': ('.rdp',),
    'application/x-redhat-package-manager': ('.rpm',),
    'application/x-rss+xml': ('.rss',),
    'application/x-ruby': ('.rb',),
    'application/x-scilab': ('.sci', '.sce',),
    'application/x-scilab-xcos': ('.xcos',),
    'application/x-sh': ('.sh',),
    'application/x-shar': ('.shar',),
    'application/x-shockwave-flash': ('.swf', '.swfl',),
    'application/x-silverlight': ('.scr',),
    'application/x-sql': ('.sql',),
    'application/x-stuffit': ('.sit', '.sitx',),
    'application/x-sv4cpio': ('.sv4cpio',),
    'application/x-sv4crc': ('.sv4crc',),
    'application/x-tar': ('.tar',),
    'application/x-tcl': ('.tcl',),
    'application/x-tex-gf': ('.gf',),
    'application/x-tex-pk': ('.pk',),
    'application/x-texinfo': ('.texinfo', '.texi',),
    'application/x-trash': ('.~', '.%', '.bak', '.old', '.sik',),
    'application/x-troff': ('.t', '.tr', '.roff',),
    'application/x-troff-man': ('.man',),
    'application/x-troff-me': ('.me',),
    'application/x-troff-ms': ('.ms',),
    'application/x-ustar': ('.ustar',),
    'application/x-wais-source': ('.src',),
    'application/x-wingz': ('.wz',),
    'application/x-x509-ca-cert': ('.crt',),
    'application/x-xcf': ('.xcf',),
    'application/x-xfig': ('.fig',),
    'application/x-xpinstall': ('.xpi',),
    'application/x-xz': ('.xz',),
    'audio/amr': ('.amr',),
    'audio/amr-wb': ('.awb',),
    'audio/annodex': ('.axa',),
    'audio/basic': ('.au', '.snd',),
    'audio/csound': ('.csd', '.orc', '.sco',),
    'audio/flac': ('.flac',),
    'audio/midi': ('.mid', '.midi', '.kar',),
    'audio/mpeg': ('.mpga', '.mpega', '.mp2', '.mp3', '.m4a',),
    'audio/mpegurl': ('.m3u',),
    'audio/ogg': ('.oga', '.ogg', '.opus', '.spx',),
    'audio/prs.sid': ('.sid',),
    'audio/x-aiff': ('.aif', '.aiff', '.aifc',),
    'audio/x-gsm': ('.gsm',),
    'audio/x-mpegurl': ('.m3u',),
    'audio/x-ms-wma': ('.wma',),
    'audio/x-ms-wax': ('.wax',),
    'audio/x-pn-realaudio': ('.ra', '.rm', '.ram',),
    'audio/x-realaudio': ('.ra',),
    'audio/x-scpls': ('.pls',),
    'audio/x-sd2': ('.sd2',),
    'audio/x-wav': ('.wav',),
    'chemical/x-alchemy': ('.alc',),
    'chemical/x-cache': ('.cac', '.cache',),
    'chemical/x-cache-csf': ('.csf',),
    'chemical/x-cactvs-binary': ('.cbin', '.cascii', '.ctab',),
    'chemical/x-cdx': ('.cdx',),
    'chemical/x-cerius': ('.cer',),
    'chemical/x-chem3d': ('.c3d',),
    'chemical/x-chemdraw': ('.chm',),
    'chemical/x-cif': ('.cif',),
    'chemical/x-cmdf': ('.cmdf',),
    'chemical/x-cml': ('.cml',),
    'chemical/x-compass': ('.cpa',),
    'chemical/x-crossfire': ('.bsd',),
    'chemical/x-csml': ('.csml', '.csm',),
    'chemical/x-ctx': ('.ctx',),
    'chemical/x-cxf': ('.cxf', '.cef',),
    'chemical/x-embl-dl-nucleotide': ('.emb', '.embl',),
    'chemical/x-galactic-spc': ('.spc',),
    'chemical/x-gamess-input': ('.inp', '.gam', '.gamin',),
    'chemical/x-gaussian-checkpoint': ('.fch', '.fchk',),
    'chemical/x-gaussian-cube': ('.cub',),
    'chemical/x-gaussian-input': ('.gau', '.gjc', '.gjf',),
    'chemical/x-gaussian-log': ('.gal',),
    'chemical/x-gcg8-sequence': ('.gcg',),
    'chemical/x-genbank': ('.gen',),
    'chemical/x-hin': ('.hin',),
    'chemical/x-isostar': ('.istr', '.ist',),
    'chemical/x-jcamp-dx': ('.jdx', '.dx',),
    'chemical/x-kinemage': ('.kin',),
    'chemical/x-macmolecule': ('.mcm',),
    'chemical/x-macromodel-input': ('.mmd', '.mmod',),
    'chemical/x-mdl-molfile': ('.mol',),
    'chemical/x-mdl-rdfile': ('.rd',),
    'chemical/x-mdl-rxnfile': ('.rxn',),
    'chemical/x-mdl-sdfile': ('.sd', '.sdf',),
    'chemical/x-mdl-tgf': ('.tgf',),
    'chemical/x-mmcif': ('.mcif',),
    'chemical/x-mol2': ('.mol2',),
    'chemical/x-molconn-Z': ('.b',),
    'chemical/x-mopac-graph': ('.gpt',),
    'chemical/x-mopac-input': ('.mop', '.mopcrt', '.mpc', '.zmt',),
    'chemical/x-mopac-out': ('.moo',),
    'chemical/x-mopac-vib': ('.mvb',),
    'chemical/x-ncbi-asn1': ('.asn',),
    'chemical/x-ncbi-asn1-ascii': ('.prt', '.ent',),
    'chemical/x-ncbi-asn1-binary': ('.val', '.aso',),
    'chemical/x-ncbi-asn1-spec': ('.asn',),
    'chemical/x-pdb': ('.pdb', '.ent',),
    'chemical/x-rosdal': ('.ros',),
    'chemical/x-swissprot': ('.sw',),
    'chemical/x-vamas-iso14976': ('.vms',),
    'chemical/x-vmd': ('.vmd',),
    'chemical/x-xtel': ('.xtel',),
    'chemical/x-xyz': ('.xyz',),
    'image/gif': ('.gif',),
    'image/ief': ('.ief',),
    'image/jp2': ('.jp2', '.jpg2',),
    'image/jpeg': ('.jpg', '.jpeg', '.jpe',),
    'image/jpm': ('.jpm',),
    'image/jpx': ('.jpx', '.jpf',),
    'image/pcx': ('.pcx',),
    'image/png': ('.png',),
    'image/svg+xml': ('.svg', '.svgz',),
    'image/tiff': ('.tiff', '.tif',),
    'image/vnd.djvu': ('.djvu', '.djv',),
    'image/vnd.microsoft.icon': ('.ico',),
    'image/vnd.wap.wbmp': ('.wbmp',),
    'image/x-canon-cr2': ('.cr2',),
    'image/x-canon-crw': ('.crw',),
    'image/x-cmu-raster': ('.ras',),
    'image/x-coreldraw': ('.cdr',),
    'image/x-coreldrawpattern': ('.pat',),
    'image/x-coreldrawtemplate': ('.cdt',),
    'image/x-corelphotopaint': ('.cpt',),
    'image/x-epson-erf': ('.erf',),
    'image/x-jg': ('.art',),
    'image/x-jng': ('.jng',),
    'image/x-ms-bmp': ('.bmp',),
    'image/x-nikon-nef': ('.nef',),
    'image/x-olympus-orf': ('.orf',),
    'image/x-photoshop': ('.psd',),
    'image/x-portable-anymap': ('.pnm',),
    'image/x-portable-bitmap': ('.pbm',),
    'image/x-portable-graymap': ('.pgm',),
    'image/x-portable-pixmap': ('.ppm',),
    'image/x-rgb': ('.rgb',),
    'image/x-xbitmap': ('.xbm',),
    'image/x-xpixmap': ('.xpm',),
    'image/x-xwindowdump': ('.xwd',),
    'message/rfc822': ('.eml',),
    'model/iges': ('.igs', '.iges',),
    'model/mesh': ('.msh', '.mesh', '.silo',),
    'model/vrml': ('.wrl', '.vrml',),
    'model/x3d+vrml': ('.x3dv',),
    'model/x3d+xml': ('.x3d',),
    'model/x3d+binary': ('.x3db',),
    'text/cache-manifest': ('.appcache',),
    'text/calendar': ('.ics', '.icz',),
    'text/css': ('.css',),
    'text/csv': ('.csv',),
    'text/h323': ('.323',),
    'text/html': ('.html', '.htm', '.shtml',),
    'text/iuls': ('.uls',),
    'text/mathml': ('.mml',),
    'text/plain': ('.asc', '.txt', '.text', '.pot', '.brf', '.srt',),
    'text/richtext': ('.rtx',),
    'text/scriptlet': ('.sct', '.wsc',),
    'text/texmacs': ('.tm',),
    'text/tab-separated-values': ('.tsv',),
    'text/turtle': ('.ttl',),
    'text/vcard': ('.vcf', '.vcard',),
    'text/vnd.sun.j2me.app-descriptor': ('.jad',),
    'text/vnd.wap.wml': ('.wml',),
    'text/vnd.wap.wmlscript': ('.wmls',),
    'text/x-bibtex': ('.bib',),
    'text/x-boo': ('.boo',),
    'text/x-c++hdr': ('.h++', '.hpp', '.hxx', '.hh',),
    'text/x-c++src': ('.c++', '.cpp', '.cxx', '.cc',),
    'text/x-chdr': ('.h',),
    'text/x-component': ('.htc',),
    'text/x-csh': ('.csh',),
    'text/x-csrc': ('.c',),
    'text/x-dsrc': ('.d',),
    'text/x-diff': ('.diff', '.patch',),
    'text/x-haskell': ('.hs',),
    'text/x-java': ('.java',),
    'text/x-lilypond': ('.ly',),
    'text/x-literate-haskell': ('.lhs',),
    'text/x-moc': ('.moc',),
    'text/x-pascal': ('.p', '.pas',),
    'text/x-pcs-gcd': ('.gcd',),
    'text/x-perl': ('.pl', '.pm',),
    'text/x-python': ('.py',),
    'text/x-scala': ('.scala',),
    'text/x-setext': ('.etx',),
    'text/x-sfv': ('.sfv',),
    'text/x-sh': ('.sh',),
    'text/x-tcl': ('.tcl', '.tk',),
    'text/x-tex': ('.tex', '.ltx', '.sty', '.cls',),
    'text/x-vcalendar': ('.vcs',),
    'video/3gpp': ('.3gp',),
    'video/annodex': ('.axv',),
    'video/dl': ('.dl',),
    'video/dv': ('.dif', '.dv',),
    'video/fli': ('.fli',),
    'video/gl': ('.gl',),
    'video/mpeg': ('.mpeg', '.mpg', '.mpe',),
    'video/MP2T': ('.ts',),
    'video/mp4': ('.mp4',),
    'video/quicktime': ('.qt', '.mov',),
    'video/ogg': ('.ogv',),
    'video/webm': ('.webm',),
    'video/vnd.mpegurl': ('.mxu',),
    'video/x-flv': ('.flv',),
    'video/x-la-asf': ('.lsf', '.lsx',),
    'video/x-mng': ('.mng',),
    'video/x-ms-asf': ('.asf', '.asx',),
    'video/x-ms-wm': ('.wm',),
    'video/x-ms-wmv': ('.wmv',),
    'video/x-ms-wmx': ('.wmx',),
    'video/x-ms-wvx': ('.wvx',),
    'video/x-msvideo': ('.avi',),
    'video/x-sgi-movie': ('.movie',),
    'video/x-matroska': ('.mpv', '.mkv',),
    'x-conference/x-cooltalk': ('.ice',),
    'x-epoc/x-sisx-app': ('.sisx',),
    'x-world/x-vrml': ('.vrm', '.vrml', '.wrl',),
}


_db_reverse = {e: k for k, v in _db.items() for e in v}


if __name__ == '__main__':  # pragma: no cover
    print(guess_extension('image/jpeg'))
    print(guess_type('a' + guess_extension('image/jpeg')))




from os.path import join
import argparse

from alembic.config import main as alembic_main

from restfulpy.cli.base import Launcher


class MigrateLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        migrate_parser = subparsers.add_parser('migrate', help="Executes the alembic command")
        migrate_parser.add_argument('alembic_args', nargs=argparse.REMAINDER,
                                    help="For more information, please see `alembic --help`")
        return migrate_parser

    def launch(self):
        alembic_ini = join(self.args.application.root_path, 'alembic.ini')
        alembic_main(argv=['--config', alembic_ini] + self.args.alembic_args)




import os
from os.path import exists
from urllib.parse import urlparse

from sqlalchemy import create_engine
from nanohttp import settings


class AbstractDatabaseManager(object):

    def __init__(self):
        self.db_uri = settings.db.uri
        self.db_name = urlparse(self.db_uri).path.lstrip('/')
        self.admin_uri = settings.db.administrative_uri
        self.admin_db_name = urlparse(settings.db.administrative_uri).path.lstrip('/')

    def __enter__(self):
        self.engine = create_engine(self.admin_uri)
        self.connection = self.engine.connect()
        self.connection.execute('commit')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connection.close()
        self.engine.dispose()

    def create_database_if_not_exists(self):
        if not self.database_exists():
            self.create_database()

    def database_exists(self):  # pragma: no cover
        raise NotImplementedError()

    def create_database(self):  # pragma: no cover
        raise NotImplementedError()

    def drop_database(self):  # pragma: no cover
        raise NotImplementedError()


class PostgresManager(AbstractDatabaseManager):

    def __enter__(self):
        super().__enter__()
        self.connection.execute('commit')
        return self

    def database_exists(self):
        r = self.connection.execute('SELECT 1 FROM pg_database WHERE datname = \'%s\'' % self.db_name)
        try:
            ret = r.cursor.fetchall()
            return ret
        finally:
            r.cursor.close()

    def create_database(self):
        self.connection.execute('CREATE DATABASE %s' % self.db_name)
        self.connection.execute('commit')

    def drop_database(self):
        self.connection.execute('DROP DATABASE IF EXISTS %s' % self.db_name)
        self.connection.execute('commit')


class SqliteManager(AbstractDatabaseManager):

    def __init__(self):
        super().__init__()
        self.filename = self.db_uri.replace('sqlite:///', '')

    def database_exists(self):
        return exists(self.filename)

    def create_database(self):
        if self.database_exists():
            raise RuntimeError('The file is already exists: %s' % self.filename)
        print('Creating: %s' % self.filename)
        open(self.filename, 'a').close()

    def drop_database(self):
        print('Removing: %s' % self.filename)
        os.remove(self.filename)


# noinspection PyAbstractClass
class DatabaseManager(AbstractDatabaseManager):

    def __new__(cls, *args, **kwargs):
        uri = settings.db.uri
        if uri.startswith('sqlite'):
            manager_class = SqliteManager
        elif uri.startswith('postgres'):
            manager_class = PostgresManager
        else:
            raise ValueError('Unsupported database uri: %s' % uri)

        return manager_class(*args, **kwargs)



from nanohttp import quickstart

from restfulpy.cli.base import Launcher


DEFAULT_ADDRESS = '8080'


class ServeLauncher(Launcher):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('serve', help="Serves the application on http.")
        parser.add_argument('-b', '--bind', default=DEFAULT_ADDRESS, metavar='{HOST:}PORT',
                            help='Bind Address. default: %s' % DEFAULT_ADDRESS)
        return parser

    def launch(self):
        host, port = self.args.bind.split(':') if ':' in self.args.bind else ('', self.args.bind)

        quickstart(
            controller=self.args.application.root,
            host=host,
            port=port
        )




from restfulpy.testing.constants import As
from restfulpy.testing.helpers import MockupMessenger
from restfulpy.testing.documentation import DocumentaryTestApp, FormParameter
from restfulpy.testing.testcases import ModelRestCrudTestCase, WebAppTestCase



from os import makedirs
from os.path import join, exists, dirname, basename
from urllib.parse import quote, urlencode
import warnings
from pprint import pprint

import ujson
from webtest import TestApp

from restfulpy.testing.constants import DOC_HEADER, DOC_LEGEND


class RequestSignature(object):
    def __init__(self, role, method, url, query_string=None, request_headers=None, response_headers=None):
        self.role = role
        self.method = method
        self.url = url
        self.query_string = query_string
        self.request_headers = request_headers
        self.response_headers = response_headers

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        return hash((
            self.role, self.method, self.url, self.query_string,
            tuple(self.request_headers.items()) if self.request_headers else None,
            tuple(self.response_headers.items()) if self.response_headers else None
        ))


class FormParameter(object):
    def __init__(self, name, value=None, type_=str, optional=False, default=None):
        self.name = name
        self.type_ = type_
        self.optional = optional
        self.value = value
        self.default = default

    @property
    def type_string(self):
        if self.type_ is None:
            return ''
        return self.type_ if isinstance(self.type_, str) else self.type_.__name__

    @property
    def value_string(self):
        if self.value is None:
            return ''

        if self.type_ == 'file':
            return basename(self.value)

        if self.type_ is bool:
            return str(self.value).lower()

        return self.value

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        return hash(self.name)


class DocumentaryTestApp(TestApp):
    _files = []
    _signatures = set()
    _jwt_token = None
    __jwt_header_key__ = 'HTTP_AUTHORIZATION'

    def __init__(self, destination_dir, application, *args, **kwargs):
        self.application = application
        self.destination_dir = destination_dir
        super(DocumentaryTestApp, self).__init__(application, *args, **kwargs)

    @property
    def legend_filename(self):
        return join(self.destination_dir, 'legend.md')

    @property
    def jwt_token(self):
        return self._jwt_token

    @jwt_token.setter
    def jwt_token(self, value):
        self._jwt_token = value

        if value is not None:
            self.extra_environ.update({
                self.__jwt_header_key__: value,
            })
        else:
            if self.__jwt_header_key__ in self.extra_environ:
                del self.extra_environ[self.__jwt_header_key__]

    def _ensure_file(self, filename, entity):
        d = dirname(filename)
        if not exists(d):
            makedirs(d)

        if not exists(self.legend_filename):
            with open(self.legend_filename, 'w') as f:
                f.write(DOC_LEGEND)

        if filename in self._files:
            return open(filename, 'a')
        else:
            f = open(filename, 'w')
            f.write(DOC_HEADER % dict(version=self.application.version))
            f.write('\n%s' % entity)
            f.write('\n%s\n' % ('=' * len(entity)))
            self._files.append(filename)
            return f

    def document(self, role, method, url, resp, request_headers, model=None, params=None, query_string=None,
                 url_params=None):
        signature = RequestSignature(
            role, method, url,
            query_string=tuple(query_string.keys()) if query_string else None,
            request_headers=request_headers,
            response_headers=resp.headers
        )
        if signature in self._signatures:
            return
        path_parts = url.split('?')[0].split('/')[1:]
        if len(path_parts) == 1:
            p = path_parts[0].strip()
            entity = filename = p if p else 'index'
        else:
            version, entity = path_parts[0:2]
            filename = '-'.join(path_parts[:2] + [method.lower()])

        filename = join(self.destination_dir, '%s.md' % filename)
        f = self._ensure_file(filename, entity)

        # Extracting more params & info from model if available
        if params and model:
            for c in model.iter_json_columns(relationships=False, include_readonly_columns=False):
                json_name = c.info.get('json', c.key)
                column = model.get_column(c)

                if hasattr(column, 'default') and column.default:
                    default_ = column.default.arg if column.default.is_scalar else 'function(...)'
                else:
                    default_ = ''

                type_ = column.type.python_type

                if json_name in params:
                    param = params[params.index(json_name)]
                    param.default = default_
                    if param.type_ is None:
                        param.type_ = type_
                    if param.optional is None:
                        param.optional = column.nullable
                else:
                    param = FormParameter(
                        json_name,
                        type_=type_,
                        optional=column.nullable,
                        default=default_)
                    params.append(param)

        try:
            f.write('## %s `%s`\n\n' % (method.upper(), url.replace('%(', '{').replace(')s', '}')))
            f.write('Role: %s\n\n' % role)

            if url_params:
                f.write('### Url Parameters:\n\n')
                f.write('| Parameter | Value |\n')
                f.write('| --------- | ----- |\n')
                for name, value in url_params.items():
                    f.write('| %s | %s |\n' % (name, value))

            if isinstance(params, dict):
                f.write('### Request: JSON\n\n')
                f.write('```json\n')
                pprint(params, stream=f, indent=8)
                f.write('```\n')

            elif isinstance(params, list):
                f.write('### Request: Form\n\n')
                f.write('| Parameter | Optional | Type | Default | Example |\n')
                f.write('| --------- | -------- | ---- | ------- | ------- |\n')
                for param in params:
                    f.write('| %s | %s | %s | %s | %s |\n' % (
                        param.name,
                        True if method.lower() == 'put' else param.optional,
                        param.type_string,
                        param.default if param.default is not None else '',
                        param.value_string))

            if query_string:
                f.write('### Query String:\n\n')

                f.write('| Parameter | Example |\n')
                f.write('| --------- | ------- |\n')
                for name, value in query_string.items():
                    f.write('| %s | %s |\n' % (
                        name,
                        str(value)))

            if request_headers:
                f.write('### Request Headers:\n\n')
                f.write('```\n')
                for k, v in request_headers.items():
                    f.write('%s: %s\n' % (k, v))
                f.write('```\n')
            f.write('\n\n')

            f.write('### Status:\n\n')
            f.write('`%s`\n\n' % resp.status)

            f.write('### Response Headers:\n\n')
            f.write('```\n')
            for k, v in resp.headers.items():
                f.write('%s: %s\n' % (k, v))
            f.write('```\n')

            f.write('### Response Body:\n\n')
            f.write('```json\n')
            if resp.charset in ('utf8', 'utf-8'):
                for l in resp.body.decode().splitlines():
                    f.write('%s\n' % l)
            else:
                f.write('%r\n' % resp.body)
            f.write('```\n\n')
            self._signatures.add(signature)
        finally:
            f.write('\n')
            f.close()

    def send_request(self, role, method, url, query_string=None, url_params=None,
                     params=None, model=None, doc=True, json=None, content_type=None, **kwargs):
        files = []

        parameters = {}
        if json:
            kwargs.setdefault('content_type', 'application/json')
            kwargs.update(
                params=ujson.dumps(json),
                upload_files=None,
            )

        else:
            kwargs.setdefault('content_type', content_type)
            if params:
                if isinstance(params, dict):
                    parameters = params
                    if doc:
                        warnings.warn(
                            'Skipping documentation generation, because the passed parameters are plain dict.',
                            stacklevel=4)
                        doc = False
                else:
                    for param in params:
                        if param.type_ == 'file':
                            files.append((param.name, param.value))
                        else:
                            parameters[param.name] = param.value

                if parameters:
                    kwargs['params'] = parameters

            if files:
                kwargs['upload_files'] = files

        real_url = (url % url_params) if url_params else url
        real_url = quote(real_url)

        if query_string:
            real_url = '%s?%s' % (real_url, urlencode(query_string))

        resp = self._gen_request(
            method.upper(), real_url, expect_errors=True, **kwargs
        )

        kwargs.setdefault('headers', {})
        content_type = kwargs.get('content_type', content_type)
        if content_type:
            kwargs['headers']['Content-Type'] = content_type

        if doc:
            self.document(role, method, url, resp, kwargs['headers'], model=model, params=json or params,
                          query_string=query_string, url_params=url_params)
        return resp




from nanohttp import configure as nanohttp_configure, settings


__builtin_config = """

debug: true
pretty_json: true

db:
  uri: sqlite:///%(data_dir)s/devdata.db
  # uri: postgresql://postgres:postgres@localhost/restfulpy_demo_dev
  # administrative_uri: postgresql://postgres:postgres@localhost/postgres
  # test_uri: postgresql://postgres:postgres@localhost/restfulpy_test
  echo: false

migration:
  directory: %(root_path)s/migration
  ini: %(root_path)s/alembic.ini

jwt:
  secret: JWT-SECRET
  algorithm: HS256
  max_age: 86400  # 24 Hours
  refresh_token:
    secret: JWT-REFRESH-SECRET
    algorithm: HS256
    max_age: 2678400  # 30 Days

messaging:
  # default_messenger: restfulpy.messaging.providers.SmtpProvider
  default_messenger: restfulpy.messaging.ConsoleMessenger
  default_sender: restfulpy
  mako_modules_directory: %(data_dir)s/mako_modules
  template_dirs:
    - %(restfulpy_dir)s/messaging/templates

authentication:
  redis:
    host: localhost
    port: 6379
    password: ~
    db: 0

worker:
  gap: .5
  number_of_threads: 1

api_documents:
  directory: %(data_dir)s/api-documents

smtp:
  host: smtp.example.com
  port: 587
  username: user@example.com
  password: password
  local_hostname: localhost
  tls: true
  auth: true
  ssl: false

logging:
  loggers:

    default:
      handlers:
        - console
        - main
        - error
      level: debug
      formatter: default
      propagate: true
      
    root:
      level: debug
      formatter: default    
    
  handlers:

    default:
      level: notset
      max_bytes: 52428800
      formatter: default

    console:
      type: console

    main:
      type: file
      filename: %(data_dir)s/logs/%(process_name)s.log

    error:
      type: file
      level: error
      filename: %(data_dir)s/logs/%(process_name)s-error.log

  formatters:
    default:
      format: "%%(asctime)s - %%(name)s - %%(levelname)s - %%(message)s"
      date_format: "%%Y-%%m-%%d %%H:%%M:%%S"

"""


def configure(config=None, directories=None, files=None, context=None, force=False):  # pragma: no cover

    nanohttp_configure(init_value=__builtin_config, context=context, force=force)

    if config:
        settings.merge(config)

    if directories:
        settings.load_dirs(directories)

    if files:
        settings.load_files(files)



import argparse
import sys

import argcomplete

from restfulpy.cli.admin import AdminLauncher
from restfulpy.cli.base import Launcher
from restfulpy.cli.migrate import MigrateLauncher
from restfulpy.cli.serve import ServeLauncher
from restfulpy.cli.worker import WorkerLauncher
from restfulpy.cli.dev import DevLauncher


class MainLauncher(Launcher):

    def __init__(self, application):
        self.application = application
        self.parser = parser = argparse.ArgumentParser(description='%s command line interface.' % application.name)
        parser.add_argument('-p', '--process-name', metavar="PREFIX", default=application.name,
                            help='A string indicates the logger prefix for this process, it helps to configure '
                                 'separate log files per process.')
        parser.add_argument('-c', '--config-file', metavar="FILE",
                            help='List of configuration files separated by space. Default: ""')
        parser.add_argument('-d', '--config-dir', metavar="DIR",
                            help='List of configuration directories separated by space. Default: ""')
        subparsers = parser.add_subparsers(title="sub commands", prog=application.name, dest="command")

        AdminLauncher.register(subparsers)
        ServeLauncher.register(subparsers)
        MigrateLauncher.register(subparsers)
        WorkerLauncher.register(subparsers)
        DevLauncher.register(subparsers)
        application.register_cli_launchers(subparsers)
        argcomplete.autocomplete(parser)

    def launch(self, args=None):

        if args:
            cli_args = self.parser.parse_args(args)
        else:
            cli_args = self.parser.parse_args()

        cli_args.application = self.application
        self.application.process_name = cli_args.process_name
        self.application.configure(files=cli_args.config_file)
        self.application.initialize_models()
        cli_args.func(cli_args)
        sys.exit(0)

    @classmethod
    def create_parser(cls, subparsers):
        """
        Do nothing here
        """
        pass



import functools
import re
from itertools import chain

from nanohttp import context
from nanohttp.exceptions import HttpBadRequest


class FormValidator:
    def __init__(
            self,
            blacklist=None,
            exclude=None,
            filter_=None,
            whitelist=None,
            requires=None,
            exact=None,
            types=None,
            pattern=None,
            **rules_per_role
    ):
        self.default_rules = {}
        if blacklist:
            self.default_rules['blacklist'] = set(blacklist)

        if exclude:
            self.default_rules['exclude'] = set(exclude)

        if filter_:
            self.default_rules['filter_'] = set(filter_)

        if whitelist:
            self.default_rules['whitelist'] = set(whitelist)

        if requires:
            self.default_rules['requires'] = set(requires)

        if exact:
            self.default_rules['exact'] = set(exact)

        if types:
            self.default_rules['types'] = types

        if pattern:
            self.default_rules['pattern'] = pattern

        self._rules_per_role = rules_per_role

    def extract_rulesets(self, rule_name, user_rules):
        return [ruleset[rule_name] for ruleset in ([self.default_rules] + user_rules) if rule_name in ruleset]

    def extract_rules_odd(self, rule_name, user_rules) -> set:
        return set(chain(
            *self.extract_rulesets(rule_name, user_rules)
        ))

    def extract_rules_pair(self, rule_name, user_rules) -> dict:
        rulesets = self.extract_rulesets(rule_name, user_rules)
        result = {}
        for r in rulesets:
            result.update(r)
        return result

    def __call__(self, *args, **kwargs):
        input_collections = [context.form, context.query_string]
        all_input_fields = set(chain(*input_collections))
        user_rules = [v for k, v in self._rules_per_role.items() if k in context.identity.roles] \
            if context.identity else []

        denied_fields = self.extract_rules_odd('blacklist', user_rules)
        if denied_fields:
            if all_input_fields & denied_fields:
                raise HttpBadRequest('These fields are denied: [%s]' % ', '.join(denied_fields))

        excluded_fields = self.extract_rules_odd('exclude', user_rules)
        if excluded_fields:
            for collection in input_collections:
                for field in set(collection) & excluded_fields:
                    del collection[field]

        filtered_fields = self.extract_rules_odd('filter_', user_rules)
        if filtered_fields:
            for collection in input_collections:
                for field in set(collection) - filtered_fields:
                    del collection[field]

        whitelist_fields = self.extract_rules_odd('whitelist', user_rules)
        if whitelist_fields:
            if all_input_fields - whitelist_fields:
                raise HttpBadRequest(
                    'These fields are not allowed: [%s]' % ', '.join(all_input_fields - whitelist_fields)
                )

        required_fields = self.extract_rules_odd('requires', user_rules)
        if required_fields:
            if required_fields - all_input_fields:
                raise HttpBadRequest('These fields are required: [%s]' % ', '.join(required_fields - all_input_fields))

        exact_fields = self.extract_rules_odd('exact', user_rules)
        if exact_fields:
            if exact_fields != all_input_fields:
                raise HttpBadRequest('Exactly these fields are allowed: [%s]' % ', '.join(whitelist_fields))

        type_pairs = self.extract_rules_pair('types', user_rules)
        if type_pairs:
            type_keys = set(type_pairs.keys())
            for collection in input_collections:
                for field in set(collection) & type_keys:
                    desired_type = type_pairs[field]
                    try:
                        collection[field] = desired_type(collection[field])
                    except ValueError:
                        raise HttpBadRequest('The field: %s must be %s' % (field, desired_type))

        pattern_pairs = self.extract_rules_pair('pattern', user_rules)
        if pattern_pairs:
            pattern_keys = set(pattern_pairs.keys())
            for collection in input_collections:
                for field in set(collection) & pattern_keys:
                    desired_pattern = pattern_pairs[field]
                    pattern = re.compile(desired_pattern) if isinstance(desired_pattern, str) else desired_pattern
                    if pattern.match(collection[field]) is None:
                        raise HttpBadRequest(
                            'The field %s: %s must be matched with %s pattern' %
                            (field, collection[field], pattern.pattern)
                        )
        return args, kwargs


def validate_form(blacklist=None, exclude=None, filter_=None, whitelist=None, requires=None, exact=None, types=None,
                  pattern=None, **rules_per_role):
    """Creates a validation decorator based on given rules:

    :param blacklist: A list fields to raise :class:`nanohttp.exceptions.HttpBadRequest` if exists in request.
    :param exclude: A list of fields to remove from the request payload if exists.
    :param filter_: A list of fields to filter the request payload.
    :param whitelist: A list of fields to raise :class:`nanohttp.exceptions.HttpBadRequest` if anythings else found in
                      the request payload.
    :param requires: A list of fields to raise :class:`nanohttp.exceptions.HttpBadRequest` if the given fields are not
                     in the request payload.
    :param exact: A list of fields to raise :class:`nanohttp.exceptions.HttpBadRequest` if the given fields are not
                  exact match.
    :param types: A dictionary of fields and their expected types. Fields will be casted to expected types if possible.
                  Otherwise :class:`nanohttp.exceptions.HttpBadRequest` will be raised.
    :param pattern: A dictionary of fields and their expected regex patterns. Fields will be matched by expected pattern
                    if possible. Otherwise :class:`nanohttp.exceptions.HttpBadRequest` will be raised.

    :param rules_per_role: A dictionary ``{ role: { ... } }``, which you can apply above rules to single role.

    :return: A validation decorator.
    """

    def decorator(func):
        validator = FormValidator(blacklist=blacklist, exclude=exclude, filter_=filter_, whitelist=whitelist,
                                  requires=requires, exact=exact, types=types, pattern=pattern, **rules_per_role)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            args, kwargs = validator(*args, **kwargs)
            return func(*args, **kwargs)

        return wrapper

    return decorator




from restfulpy.cli.base import RequireSubCommand, Launcher
from restfulpy.db import DatabaseManager
from restfulpy.orm import setup_schema


class BasedataLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        return subparsers.add_parser('base-data', help="Setup the server's database.")

    def launch(self):
        self.args.application.insert_basedata()


class MockupDataLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        return subparsers.add_parser('mockup-data', help="Insert mockup data.")

    def launch(self):
        self.args.application.insert_mockup()


class CreateDatabaseLauncher(Launcher):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('create-db', help="Create the server's database.")
        parser.add_argument('-d', '--drop', dest='drop_db', action='store_true', default=False,
                            help='Drop existing database before create another one.')
        parser.add_argument('-s', '--schema', dest='schema', action='store_true', default=False,
                            help='Creates database schema after creating the database.')
        parser.add_argument('-b', '--basedata', action='store_true', default=False,
                            help='Implies `(-s|--schema)`, Inserts basedata after schema generation.')
        parser.add_argument('-m', '--mockup', action='store_true', default=False,
                            help='Implies `(-s|--schema, -b|--basedata)`, Inserts mockup data.')
        return parser

    def launch(self):
        with DatabaseManager() as db_admin:
            if self.args.drop_db:
                db_admin.drop_database()
            db_admin.create_database()
            if self.args.schema or self.args.basedata or self.args.mockup:
                setup_schema()
                if self.args.basedata:
                    self.args.application.insert_basedata()
                if self.args.mockup:
                    self.args.application.insert_mockup()


class DropDatabaseLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        return subparsers.add_parser('drop-db', help="Setup the server's database.")

    def launch(self):
        with DatabaseManager() as db_admin:
            db_admin.drop_database()


class SetupDatabaseLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        return subparsers.add_parser('setup-db', help="Setup the server's database.")

    def launch(self):
        setup_schema()


class AdminLauncher(Launcher, RequireSubCommand):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('admin', help="Administrate the application")
        admin_subparsers = parser.add_subparsers(title="admin command", dest="admin_command")
        SetupDatabaseLauncher.register(admin_subparsers)
        BasedataLauncher.register(admin_subparsers)
        MockupDataLauncher.register(admin_subparsers)
        DropDatabaseLauncher.register(admin_subparsers)
        CreateDatabaseLauncher.register(admin_subparsers)
        return parser



import unittest

from sqlalchemy import Unicode
from nanohttp import settings

from restfulpy.orm import DeclarativeBase, DBSession, Field, OrderableMixin
from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication


class OrderableCheckingModel(OrderableMixin, DeclarativeBase):
    __tablename__ = 'orderable_checking_model'

    title = Field(Unicode(50), primary_key=True)


class OrderableCheckingModelTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', None)
    __configuration__ = '''
    db:
      uri: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)

    def test_orderable_mixin(self):

        for i in range(3):
            # noinspection PyArgumentList
            instance = OrderableCheckingModel(
                title='test title %s' % i,
                order=i
            )
            DBSession.add(instance)
            DBSession.commit()

        instances = OrderableCheckingModel.apply_default_sort().all()
        self.assertEqual(instances[0].order, 0)
        self.assertEqual(instances[2].order, 2)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import sys
import warnings
import importlib
import importlib.util
import uuid
import re
from os.path import dirname, abspath
from datetime import datetime, timedelta
from hashlib import md5


ZERO = timedelta(0)


def import_python_module_by_filename(name, module_filename):
    """
    Import's a file as a python module, with specified name.

    Don't ask about the `name` argument, it's required.
    
    :param name: The name of the module to override upon imported filename.
    :param module_filename: The filename to import as a python module.
    :return: The newly imported python module.
    """

    sys.path.append(abspath(dirname(module_filename)))
    spec = importlib.util.spec_from_file_location(
        name,
        location=module_filename)
    imported_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(imported_module)
    return imported_module


def construct_class_by_name(name, *args, **kwargs):
    """
    Construct a class by module path name using *args and **kwargs

    Don't ask about the `name` argument, it's required.

    :param name: class name
    :return: The newly imported python module.
    """
    parts = name.split('.')
    module_name, class_name = '.'.join(parts[:-1]), parts[-1]
    module = importlib.import_module(module_name)
    return getattr(module, class_name)(*args, **kwargs)


def deprecated(func):  # pragma: no cover
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emmitted
    when the function is used.
    """

    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # Turn off filter
        warnings.warn(
            'Call to deprecated function %s.' % func.__name__,
            category=DeprecationWarning,
            stacklevel=2
        )
        warnings.simplefilter('default', DeprecationWarning)  # Reset filter
        return func(*args, **kwargs)

    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    new_func.__dict__.update(func.__dict__)
    return new_func


def format_iso_datetime(stamp):
    """ Return a string representing the date and time in ISO 8601 format.
        If the time is in UTC, adds a 'Z' directly after the time without
        a space.

        see http://en.wikipedia.org/wiki/ISO_8601.

        >>> class EET(tzinfo):
        ...     def utcoffset(self, dt):
        ...         return timedelta(minutes=120)
        ...     def dst(self, dt):
        ...         return timedelta()
        >>> format_iso_datetime(datetime(2012, 2, 22, 12, 52, 29, 300))
        '2012-02-22T12:52:29'
        >>> format_iso_datetime(datetime(2012, 2, 22, 12, 52, 29, 300,
        ...     tzinfo=UTC))
        '2012-02-22T12:52:29Z'
        >>> format_iso_datetime(datetime(2012, 2, 22, 12, 52, 29, 300,
        ...     tzinfo=EET()))
        '2012-02-22T12:52:29+02:00'
    """
    if stamp.tzinfo:
        if stamp.utcoffset() == ZERO:
            return datetime(*stamp.timetuple()[:6]).isoformat() + 'Z'
    if stamp.microsecond:
        stamp = stamp.replace(microsecond=0)
    return stamp.isoformat()


def format_iso_time(stamp):
    """ Return a string representing the time in ISO 8601 format.
        If the time is in UTC, adds a 'Z' directly after the time without
        a space.

        see http://en.wikipedia.org/wiki/ISO_8601.

        >>> class EET(tzinfo):
        ...     def utcoffset(self, dt):
        ...         return timedelta(minutes=120)
        ...     def dst(self, dt):
        ...         return timedelta()
        >>> format_iso_time(time(12, 52, 29, 300))
        '12:52:29'
        >>> format_iso_time(time(12, 52, 29, 300,
        ...     tzinfo=UTC))
        '12:52:29Z'
        >>> format_iso_time(time(12, 52, 29, 300,
        ...     tzinfo=EET()))
        '12:52:29+02:00'
    """
    if stamp.microsecond:
        stamp = stamp.replace(microsecond=0)
    if stamp.tzinfo:
        if stamp.utcoffset() == ZERO:
            return stamp.replace(tzinfo=None).isoformat() + 'Z'
        else:
            return stamp.isoformat()
    else:
        return stamp.isoformat()


def random_password(length=10):
    """Returns a random string of length string_length."""
    random = str(uuid.uuid4())  # Convert UUID format to a Python string.
    random = random.upper()  # Make all characters uppercase.
    random = random.replace("-", "")  # Remove the UUID '-'.
    return random[0:length]  # Return the random string.


def to_camel_case(text):
    return re.sub("(_\w)", lambda x: x.group(1)[1:].upper(), text)


def copy_stream(source, target, *, chunk_size: int= 16 * 1024) -> int:
    length = 0
    while 1:
        buf = source.read(chunk_size)
        if not buf:
            break
        length += len(buf)
        target.write(buf)
    return length


def md5sum(f):
    if isinstance(f, str):
        file_obj = open(f, 'rb')
    else:
        file_obj = f

    try:
        checksum = md5()
        while True:
            d = file_obj.read(1024)
            if not d:
                break
            checksum.update(d)
        return checksum.digest()
    finally:
        if file_obj is not f:
            file_obj.close()



import copy
import unittest

from nanohttp import context, json, settings

from restfulpy.principal import DummyIdentity
from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication
from restfulpy.validation import validate_form
from restfulpy.controllers import RestController, RootController


class ValidationBlackListController(RestController):
    @json
    @validate_form(blacklist=['blacklistParamForAll'],
                   client={'blacklist': ['blacklistParamForClient']},
                   admin={'blacklist': ['blacklistParamForAdmin']})
    def post(self):
        result = copy.deepcopy(context.form)
        result.update(context.query_string)
        return result


class Root(RootController):
    validation = ValidationBlackListController()


class ValidationBlackListTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', Root())

    @classmethod
    def configure_app(cls):
        super().configure_app()
        settings.merge("""
            logging:
              loggers:
                default:
                  level: info
            """)

    def test_validation_blacklist(self):
        # Test `blacklist`
        # role -> All
        self.wsgi_app.jwt_token = DummyIdentity().dump().decode()
        self.request('All', 'POST', '/validation', doc=False)
        self.request('All', 'POST', '/validation', doc=False, params={'customParam': 'param'})
        self.request('All', 'POST', '/validation', doc=False, params={'blacklistParamForAll': 'param'},
                     expected_status=400)
        self.request('All', 'POST', '/validation', doc=False, params={'blacklistParamForClient': 'param'})
        self.request('All', 'POST', '/validation', doc=False, params={'blacklistParamForAdmin': 'param'})
        # -----------------------------
        # role -> Client
        self.wsgi_app.jwt_token = DummyIdentity('client').dump().decode()
        self.request('Client', 'POST', '/validation', doc=False)
        self.request('Client', 'POST', '/validation', doc=False, params={'customParam': 'param'})
        self.request('Client', 'POST', '/validation', doc=False, params={'blacklistParamForAll': 'param'},
                     expected_status=400)
        self.request(
            'Client', 'POST', '/validation', doc=False,
            params={'blacklistParamForClient': 'param'},
            expected_status=400
        )
        self.request('Client', 'POST', '/validation', doc=False, params={
            'blacklistParamForAdmin': 'param'
        })
        # -----------------------------
        # role -> Admin
        self.wsgi_app.jwt_token = DummyIdentity('admin').dump().decode()
        self.request('Admin', 'POST', '/validation', doc=False)
        self.request('Admin', 'POST', '/validation', doc=False, params={'customParam': 'param'})
        self.request('Admin', 'POST', '/validation', doc=False, params={'blacklistParamForAll': 'param'},
                     expected_status=400)
        self.request('Admin', 'POST', '/validation', doc=False, params={'blacklistParamForClient': 'param'})
        self.request('Admin', 'POST', '/validation', doc=False, params={'blacklistParamForAdmin': 'param'},
                     expected_status=400)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




import sys
import signal
import threading

from nanohttp import settings

from restfulpy.cli.base import Launcher, RequireSubCommand


class StartLauncher(Launcher):
    __command__ = 'start'

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser(cls.__command__, help='Starts the background worker.')

        parser.add_argument('-g', '--gap', type=int, default=None, help='Gap between run next task.')

        parser.add_argument('-s', '--status', default=[], action='append', help='Task status to process')

        parser.add_argument('-n', '--number-of-threads', type=int, default=None, help='Number of working threads')

        parser.add_argument(
            '-f', '--filter', default=None, type=str, action='store', help='Custom SQL filter for tasks'
        )
        return parser

    def launch(self):

        from restfulpy.taskqueue import worker

        signal.signal(signal.SIGINT, self.kill_signal_handler)
        signal.signal(signal.SIGTERM, self.kill_signal_handler)

        if not self.args.status:
            self.args.status = {'new'}

        if self.args.gap is not None:
            settings.worker.merge({'gap': self.args.gap})

        print('The following task types would be processed with gap of %ds:' % settings.worker.gap)
        print('|'.join(self.args.type) or 'all types!')
        print('Tracking task status(es): %s' % ','.join(self.args.status))

        if self.args.exclude_type:
            print('The following task types would not be processed:')
            print('|'.join(self.args.exclude_type))

        number_of_threads = self.args.number_of_threads or settings.worker.number_of_threads
        for i in range(number_of_threads):
            t = threading.Thread(
                    target=worker,
                    name='restfulpy-worker-thread-%s' % i,
                    daemon=True,
                    kwargs=dict(
                        statuses=self.args.status,
                        filters=self.args.filter
                    )
                )
            t.start()

        print('Worker started with %d threads' % number_of_threads)
        print('Press Ctrl+C to terminate worker')
        signal.pause()

    # noinspection PyUnusedLocal
    @staticmethod
    def kill_signal_handler(signal_number, frame):

        if signal_number == signal.SIGINT:
            print('You pressed Ctrl+C!')
        elif signal_number in (signal.SIGTERM, signal.SIGKILL):
            print('OS Killed Me!')

        sys.stdin.close()
        sys.stderr.close()
        sys.stdout.close()
        sys.exit(1)


class CleanupLauncher(Launcher):
    @classmethod
    def create_parser(cls, subparsers):
        return subparsers.add_parser('cleanup', help='Clean database before starting worker processes')

    def launch(self):
        from restfulpy.orm import DBSession
        from restfulpy.taskqueue import Task

        Task.cleanup(DBSession)
        DBSession.commit()


class WorkerLauncher(Launcher, RequireSubCommand):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('worker', help="Task queue administration")
        worker_subparsers = parser.add_subparsers(title="worker command", dest="worker_command")
        StartLauncher.register(worker_subparsers)
        CleanupLauncher.register(worker_subparsers)
        return parser



from restfulpy.application import Application


# noinspection PyAbstractClass
class MockupApplication(Application):
    builtin_configuration = '''
    db:
      test_uri: postgresql://postgres:postgres@localhost/restfulpy_test
      administrative_uri: postgresql://postgres:postgres@localhost/postgres      
    '''

    def configure(self, files=None, context=None, **kwargs):
        _context = dict(
            process_name='restfulpy_unittests'
        )
        if context:
            _context.update(context)
        super().configure(files=files, context=_context, **kwargs)



import unittest

from sqlalchemy import Integer, Unicode
from nanohttp import settings, HttpBadRequest
from nanohttp.contexts import Context

from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication
from restfulpy.orm import DeclarativeBase, Field, DBSession, PaginationMixin


class PagingObject(PaginationMixin, DeclarativeBase):
    __tablename__ = 'paging_object'
    __max_take__ = 4

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))


class PaginationMixinTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', None)
    __configuration__ = '''
    db:
      uri: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)

    def test_pagination_mixin(self):
        for i in range(1, 6):
            # noinspection PyArgumentList
            obj = PagingObject(
                title='object %s' % i,
            )
            DBSession.add(obj)
        DBSession.commit()

        with Context({'QUERY_STRING': 'take=2&skip=1'}, self.application) as context:
            self.assertEqual(PagingObject.paginate_by_request().count(), 2)
            self.assertEqual(context.response_headers['X-Pagination-Take'], '2')
            self.assertEqual(context.response_headers['X-Pagination-Skip'], '1')
            self.assertEqual(context.response_headers['X-Pagination-Count'], '5')

        with Context({'QUERY_STRING': 'take=two&skip=one'}, self.application):
            self.assertEqual(PagingObject.paginate_by_request().count(), 4)

        with Context({'QUERY_STRING': 'take=5'}, self.application):
            self.assertRaises(HttpBadRequest, PagingObject.paginate_by_request)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import ujson
from nanohttp import context, json, RestController, configure
from nanohttp.contexts import Context

from restfulpy.controllers import JsonPatchControllerMixin


class BiscuitsController(RestController):
    @json
    def put(self, id_: int = None):
        result = {}
        result.update(context.form)
        result['id'] = id_
        return result

    @json
    def get(self, id_: int = None):
        result = {}
        result.update(context.form)
        result['id'] = id_
        return result

    @json
    def error(self):
        raise Exception()


class SimpleJsonPatchController(JsonPatchControllerMixin, RestController):
    biscuits = BiscuitsController()


class JsonPatchTestCase(unittest.TestCase):
    __configuration__ = '''
    '''

    @classmethod
    def setUpClass(cls):
        configure(init_value=cls.__configuration__, force=True)

    def test_jsonpatch_rfc6902(self):
        environ = {
            'REQUEST_METHOD': 'PATCH'
        }
        with Context(environ):
            controller = SimpleJsonPatchController()
            context.form = [
                {"op": "put", "path": "biscuits/1", "value": {"name": "Ginger Nut"}},
                {"op": "get", "path": "biscuits/2", "value": {"name": "Ginger Nut"}},
            ]
            result = ujson.loads(controller())
            self.assertEqual(len(result), 2)

    def test_jsonpatch_error(self):
        environ = {
            'REQUEST_METHOD': 'PATCH'
        }
        with Context(environ):
            controller = SimpleJsonPatchController()
            context.form = [
                {"op": "put", "path": "biscuits/1", "value": {"name": "Ginger Nut"}},
                {"op": "error", "path": "biscuits", "value": None},
            ]
            self.assertRaises(Exception, controller)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()





class Launcher:
    no_launch = False
    parser = None

    @classmethod
    def create_parser(cls, subparsers):
        raise NotImplementedError

    @classmethod
    def register(cls, subparsers):
        parser = cls.create_parser(subparsers)
        instance = cls()
        instance.parser = parser
        if not cls.no_launch:
            parser.set_defaults(func=instance)
        return instance

    def __call__(self, *args):
        if len(args):
            self.args = args[0]
        else:
            self.args = None
        self.launch()

    def launch(self):
        if self.parser:
            self.parser.print_help()


class RequireSubCommand:
    no_launch = True



from os.path import basename
import smtplib
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from mako.lookup import TemplateLookup
from nanohttp import settings, LazyAttribute

from restfulpy.utils import construct_class_by_name


class Messenger(object):
    """
    The abstract base class for everyone messaging operations
    """

    def render_body(self, body, template_filename=None):
        if template_filename:
            mako_template = self.lookup.get_template(template_filename)
        else:
            mako_template = None

        if mako_template:
            return mako_template.render(**body)
        else:
            return body

    @LazyAttribute
    def lookup(self):
        return TemplateLookup(
            module_directory=settings.messaging.mako_modules_directory,
            directories=settings.messaging.template_dirs,
            input_encoding='utf8'
        )

    def send(self, to, subject, body, cc=None, bcc=None, template_filename=None, from_=None,
             attachments=None):  # pragma: no cover
        raise NotImplementedError


class SmtpProvider(Messenger):

    def send(self, to, subject, body, cc=None, bcc=None, template_filename=None, from_=None, attachments=None):
        """
        Sending messages with SMTP server
        """
        # FIXME: Exception handling

        body = self.render_body(body, template_filename)

        smtp_config = settings.smtp
        smtp_server = (smtplib.SMTP_SSL if smtp_config.ssl else smtplib.SMTP)(
            host=smtp_config.host,
            port=smtp_config.port,
            local_hostname=smtp_config.local_hostname
        )
        if smtp_config.tls:  # pragma: no cover
            smtp_server.starttls()

        if smtp_config.auth:  # pragma: no cover
            smtp_server.login(smtp_config.username, smtp_config.password)

        from_ = from_ or smtp_config.username

        msg = MIMEMultipart()
        msg['Subject'] = subject
        msg['From'] = from_
        msg['To'] = to
        if cc:
            msg['Cc'] = cc
        if bcc:
            msg['Bcc'] = bcc

        html_part = MIMEText(body, 'html')
        msg.attach(html_part)
        if attachments:
            for attachment in attachments:
                assert hasattr(attachment, 'name')
                attachment_part = MIMEApplication(attachment.read(), Name=basename(attachment.name))
                attachment_part['Content-Disposition'] = 'attachment; filename="%s"' % basename(attachment.name)
                msg.attach(attachment_part)

        smtp_server.send_message(msg)
        smtp_server.quit()


class ConsoleMessenger(Messenger):
    def send(self, to, subject, body, cc=None, bcc=None, template_filename=None, from_=None, attachments=None):
        """
        Sending messages by email
        """

        body = self.render_body(body, template_filename)
        print(body)


def create_messenger() -> Messenger:
    return construct_class_by_name(settings.messaging.default_messenger)



from datetime import datetime

from sqlalchemy import DateTime, Integer, between, desc
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import synonym, SynonymProperty
from sqlalchemy.sql.expression import nullslast, nullsfirst
from sqlalchemy.events import event
from nanohttp import context, HttpBadRequest, HttpConflict

from restfulpy.orm.field import Field
from restfulpy.utils import to_camel_case


class TimestampMixin:
    created_at = Field(DateTime, default=datetime.now, nullable=False, json='createdAt', readonly=True)


class ModifiedMixin(TimestampMixin):
    modified_at = Field(DateTime, nullable=True, json='modifiedAt', readonly=True)

    @property
    def last_modification_time(self):
        return self.modified_at or self.created_at

    # noinspection PyUnusedLocal
    @staticmethod
    def on_update(mapper, connection, target):
        target.modified_at = datetime.now()

    @classmethod
    def __declare_last__(cls):
        event.listen(cls, 'before_update', cls.on_update)


class OrderableMixin:
    order = Field("order", Integer, default=0, nullable=False)
    __mapper_args__ = dict(order_by=order)

    @classmethod
    def apply_default_sort(cls, query=None):
        return (query or cls.query).order_by(cls.order)


class SoftDeleteMixin:
    removed_at = Field(DateTime, nullable=True, json='removedAt', readonly=True)

    def assert_is_not_deleted(self):
        if self.is_deleted:
            raise ValueError('Object is already deleted.')

    def assert_is_deleted(self):
        if not self.is_deleted:
            raise ValueError('Object is not deleted.')

    @property
    def is_deleted(self):
        return self.removed_at is not None

    def soft_delete(self, ignore_errors=False):
        if not ignore_errors:
            self.assert_is_not_deleted()
        self.removed_at = datetime.now()

    def soft_undelete(self, ignore_errors=False):
        if not ignore_errors:
            self.assert_is_deleted()
        self.removed_at = None

    @staticmethod
    def on_delete(mapper, connection, target):
        raise HttpConflict('Cannot remove this object: %s' % target)

    @classmethod
    def __declare_last__(cls):
        event.listen(cls, 'before_delete', cls.on_delete)

    @classmethod
    def filter_deleted(cls, query=None):
        # noinspection PyUnresolvedReferences
        return (query or cls.query).filter(cls.removed_at.isnot(None))

    @classmethod
    def exclude_deleted(cls, query=None):
        # noinspection PyUnresolvedReferences
        return (query or cls.query).filter(cls.removed_at.is_(None))


class ActivationMixin:
    activated_at = Field(DateTime, nullable=True, json='activatedAt', readonly=True, protected=True)

    @hybrid_property
    def is_active(self):
        return self.activated_at is not None

    @is_active.setter
    def is_active(self, value):
        self.activated_at = datetime.now() if value else None

    @is_active.expression
    def is_active(self):
        # noinspection PyUnresolvedReferences
        return self.activated_at.isnot(None)

    @classmethod
    def filter_activated(cls, query=None):
        # noinspection PyUnresolvedReferences
        return (query or cls.query).filter(cls.is_active)

    @classmethod
    def import_value(cls, column, v):
        if column.key == cls.is_active.key and not isinstance(v, bool):
            return str(v).lower() == 'true'
        return super().import_value(column, v)


class PaginationMixin:
    __take_header_key__ = 'X_TAKE'
    __skip_header_key__ = 'X_SKIP'
    __max_take__ = 100

    @classmethod
    def paginate_by_request(cls, query=None):
        # noinspection PyUnresolvedReferences
        query = query or cls.query

        try:
            take = int(
                context.query_string.get('take') or context.environ.get(cls.__take_header_key__) or cls.__max_take__)
        except ValueError:
            take = cls.__max_take__

        try:
            skip = int(context.query_string.get('skip') or context.environ.get(cls.__skip_header_key__) or 0)
        except ValueError:
            skip = 0

        if take > cls.__max_take__:
            raise HttpBadRequest()

        context.response_headers.add_header('X-Pagination-Take', str(take))
        context.response_headers.add_header('X-Pagination-Skip', str(skip))
        context.response_headers.add_header('X-Pagination-Count', str(query.count()))
        return query.offset(skip).limit(take)  # [skip:skip + take] Commented by vahid


class FilteringMixin:
    @classmethod
    def filter_by_request(cls, query=None):
        # noinspection PyUnresolvedReferences
        query = query or cls.query

        # noinspection PyUnresolvedReferences
        for c in cls.iter_json_columns():
            json_name = c.info.get('json', to_camel_case(c.key))
            if json_name in context.query_string:
                value = context.query_string[json_name]
                query = cls._filter_by_column_value(query, c, value)

        return query

    @classmethod
    def _filter_by_column_value(cls, query, column, value):

        import_value = getattr(cls, 'import_value')
        if not isinstance(value, str):
            raise HttpBadRequest()

        if value.startswith('^') or value.startswith('!^'):
            value = value.split(',')
            not_ = value[0].startswith('!^')
            first_item = value[0][2 if not_ else 1:]
            items = [first_item] + value[1:]
            items = [i for i in items if i.strip()]
            if not len(items):
                raise HttpBadRequest('Invalid query string: %s' % value)
            expression = column.in_([import_value(column, j) for j in items])
            if not_:
                expression = ~expression

        elif value.startswith('~'):
            values = value[1:].split(',')
            start, end = [import_value(column, v) for v in values]
            expression = between(column, start, end)

        elif value == 'null':
            expression = column.is_(None)
        elif value == '!null':
            expression = column.isnot(None)
        elif value.startswith('!'):
            expression = column != import_value(column, value[1:])
        elif value.startswith('>='):
            expression = column >= import_value(column, value[2:])
        elif value.startswith('>'):
            expression = column > import_value(column, value[1:])
        elif value.startswith('<='):
            expression = column <= import_value(column, value[2:])
        elif value.startswith('<'):
            expression = column < import_value(column, value[1:])
        elif value.startswith('%~'):
            expression = column.ilike('%%%s%%' % import_value(column, value[2:]))
        elif value.startswith('%'):
            expression = column.like('%%%s%%' % import_value(column, value[1:]))
        else:
            expression = column == import_value(column, value)

        return query.filter(expression)


class OrderingMixin:
    @classmethod
    def _sort_by_key_value(cls, query, column, descending=False):
        expression = column

        if descending:
            expression = desc(expression)
            pre = nullsfirst
        else:
            pre = nullslast

        return query.order_by(pre(expression))

    @classmethod
    def sort_by_request(cls, query=None):
        # noinspection PyUnresolvedReferences
        query = query or cls.query

        sort_exp = context.query_string.get('sort', '').strip()
        if not sort_exp:
            if issubclass(cls, OrderableMixin):
                return cls.apply_default_sort(query)
            return query

        sort_columns = {c[1:] if c.startswith('-') else c: 'desc' if c.startswith('-') else 'asc'
                        for c in sort_exp.split(',')}

        criteria = cls.create_sort_criteria(sort_columns)

        for criterion in criteria:
            query = cls._sort_by_key_value(query, *criterion)

        return query


class ApproveRequiredMixin:
    approved_at = Field(DateTime, nullable=True, json='approvedAt', readonly=True)

    @hybrid_property
    def is_approved(self):
        return self.approved_at is not None

    @is_approved.setter
    def is_approved(self, value):
        self.approved_at = datetime.now() if value else None

    @is_approved.expression
    def is_approved(self):
        # noinspection PyUnresolvedReferences
        return self.approved_at.isnot(None)

    @classmethod
    def import_value(cls, column, v):
        if column.key == cls.is_approved.key and not isinstance(v, bool):
            return str(v).lower() == 'true'
        return super().import_value(column, v)

    @classmethod
    def filter_approved(cls, query=None):
        # noinspection PyUnresolvedReferences
        return (query or cls.query).filter(cls.is_approved)


class FullTextSearchMixin:
    __ts_vector__ = None

    @classmethod
    def search(cls, expressions, query=None):
        expressions = expressions.replace(' ', '|')
        query = (query or cls.query).filter(
            cls.__ts_vector__.match(expressions)
        )
        return query



import sys
import unittest

from nanohttp.contexts import context, Context

from restfulpy.principal import ImpersonateAs, DummyIdentity


class ImpersonationTest(unittest.TestCase):
    def setUp(self):
        self.http_context = Context({})
        self.http_context.__enter__()

    def tearDown(self):
        self.http_context.__exit__(*sys.exc_info())

    def test_impersonation(self):
        role1 = 'role-1'
        role2 = 'role-2'
        role3 = 'role-3'
        role4 = 'role-4'

        # Simple test
        role_1_principal = DummyIdentity(role1)
        with ImpersonateAs(role_1_principal):
            self.assertEqual(context.identity, role_1_principal)
            self.assertTrue(context.identity.is_in_roles(role1))

            # Now we change the role
            role_2_principal = DummyIdentity(role2)
            with ImpersonateAs(role_2_principal):
                self.assertFalse(context.identity.is_in_roles(role1))
                self.assertTrue(context.identity.is_in_roles(role2))

        # Multiple roles
        role_3_4_principal = DummyIdentity(role3, role4)
        with ImpersonateAs(role_3_4_principal):
            self.assertTrue(context.identity.is_in_roles(role3))
            self.assertTrue(context.identity.is_in_roles(role4))

            self.assertFalse(context.identity.is_in_roles(role1))
            self.assertFalse(context.identity.is_in_roles(role2))


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




import functools
from os.path import exists

from nanohttp import settings, context
from sqlalchemy import create_engine as sa_create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.sql.schema import MetaData
from sqlalchemy.ext.declarative import declarative_base
from alembic import config, command

from .field import Field, relationship, composite
from .mixines import ModifiedMixin, SoftDeleteMixin, TimestampMixin, ActivationMixin, PaginationMixin, FilteringMixin, \
    OrderingMixin, OrderableMixin, ApproveRequiredMixin, FullTextSearchMixin
from .metadata import MetadataField
from .models import BaseModel
from .fulltext_search import to_tsvector, fts_escape

# Global session manager: DBSession() returns the Thread-local
# session object appropriate for the current web request.
session_factory = sessionmaker(
    autoflush=False,
    autocommit=False,
    expire_on_commit=True,
    twophase=False)

DBSession = scoped_session(session_factory)

# Global metadata.
metadata = MetaData()

DeclarativeBase = declarative_base(cls=BaseModel, metadata=metadata)

# There are two convenient ways for you to spare some typing.
# You can have a query property on all your model classes by doing this:
DeclarativeBase.query = DBSession.query_property()


def create_engine(uri=None, echo=None):
    return sa_create_engine(uri or settings.db.uri, echo=echo or settings.db.echo)


def init_model(engine):
    """
    Call me before using any of the tables or classes in the model.
    :param engine: SqlAlchemy engine to bind the session
    :return:
    """

    if DBSession.registry.has():
        DBSession.remove()

    DBSession.configure(bind=engine)


def drop_all(session=None):  # pragma: no cover
    session = session or DBSession
    # noinspection PyUnresolvedReferences
    engine = session.bind
    metadata.drop_all(bind=engine)


def setup_schema(session=None):  # pragma: no cover
    session = session or DBSession
    # noinspection PyUnresolvedReferences
    engine = session.bind
    metadata.create_all(bind=engine)

    if hasattr(settings, 'migration') and exists(settings.migration.directory):
        alembic_cfg = config.Config()
        alembic_cfg.set_main_option("script_location", settings.migration.directory)
        alembic_cfg.set_main_option("sqlalchemy.url", str(engine.url))
        alembic_cfg.config_file_name = settings.migration.ini
        command.stamp(alembic_cfg, "head")


def create_thread_unsafe_session():  # pragma: no cover
    return session_factory()


def commit(func):

    # noinspection PyUnresolvedReferences
    @functools.wraps(func)
    def wrapper(*args, **kwargs):

        if hasattr(context, 'jsonpatch'):
            return func(*args, **kwargs)

        try:
            result = func(*args, **kwargs)
            DBSession.commit()
            return result
        except:
            if DBSession.is_active:
                DBSession.rollback()
            raise

    return wrapper




from os import path, makedirs
import logging
from logging.handlers import RotatingFileHandler

from nanohttp import settings, LazyAttribute


_loggers = {}
_handlers = {}
_formatters = {}
root_logger_is_configured = False


def get_level(name):
    return {
        'notset': logging.NOTSET,       # 0
        'debug': logging.DEBUG,         # 10
        'info': logging.INFO,           # 20
        'warning': logging.WARNING,     # 30
        'error': logging.ERROR,         # 40
        'critical': logging.CRITICAL    # 50
    }[name]


def ensure_formatter(name):
    if name not in _formatters:
        formatter_config = settings.logging.formatters.default.copy()
        formatter_config.update(settings.logging.formatters.get(name, {}))
        _formatters[name] = logging.Formatter(formatter_config.format, formatter_config.date_format)
    return _formatters[name]


def ensure_handler(name):
    if name not in _handlers:

        handler_config = settings.logging.handlers.default.copy()
        handler_config.update(settings.logging.handlers.get(name, {}))

        if handler_config.type == 'console':
            handler = logging.StreamHandler()
        elif handler_config.type == 'file':
            directory = path.dirname(handler_config.filename)
            if not path.exists(directory):  # pragma: no cover
                makedirs(directory)
            handler = RotatingFileHandler(
                handler_config.filename,
                encoding='utf-8',
                maxBytes=handler_config.get('max_bytes', 52428800)
            )
        else:  # pragma: no cover
            raise ValueError('Invalid handler type: %s' % handler_config.type)

        if handler_config.level != 'notset':
            handler.setLevel(get_level(handler_config.level))

        # Attaching newly created formatter to the handler
        handler.setFormatter(ensure_formatter(handler_config.formatter))
        _handlers[name] = handler

    return _handlers[name]


def ensure_root_logger():
    global root_logger_is_configured

    if root_logger_is_configured:
        return

    # Rebasing with default config
    logger_config = settings.logging.loggers.default.copy()
    logger_config.update(settings.logging.loggers.root)

    logging.basicConfig(handlers=logger_config.handlers, level=logger_config.level)
    root_logger_is_configured = True


def ensure_logger(name):
    global root_logger_is_configured
    ensure_root_logger()

    if name not in _loggers:
        # Rebasing with default config
        logger_config = settings.logging.loggers.default.copy()
        logger_config.update(settings.logging.loggers.get(name, {}))
        level = get_level(logger_config.level)

        # Creating logger
        logger = logging.getLogger(name)
        logger.setLevel(level)
        logger.propagate = logger_config.propagate

        # Creating Handlers
        for handler_name in logger_config.handlers:
            logger.addHandler(ensure_handler(handler_name))

        # Adding the first log entry
        _loggers[name] = logger

    return _loggers[name]


class LoggerProxy(object):  # pragma: no cover
    def __init__(self, name):
        self.name = name

    @LazyAttribute
    def logger(self):
        return ensure_logger(self.name)

    def info(self, *args, **kw):
        self.logger.info(*args, **kw)

    def debug(self, *args, **kw):
        self.logger.debug(*args, **kw)

    def error(self, *args, **kw):
        self.logger.error(*args, **kw)

    def warning(self, *args, **kw):
        self.logger.warning(*args, **kw)

    def critical(self, *args, **kw):
        self.logger.critical(*args, **kw)

    def exception(self, *args, **kw):
        self.logger.exception(*args, **kw)


def get_logger(logger_name='main'):
    return LoggerProxy(logger_name)




import yaml
from nanohttp import settings
from restfulpy.cli.base import RequireSubCommand, Launcher


class EmailLauncher(Launcher):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('email', help="Sends a test email")
        parser.add_argument('target')
        parser.add_argument('-m', '--message', default='The email body')
        parser.add_argument('-s', '--subject', default='Test subject')
        parser.add_argument('-f', '--from', dest='from_', default='restfulpy@example.com')
        return parser

    def launch(self):
        from restfulpy.messaging import SmtpProvider

        p = SmtpProvider()
        p.send(self.args.target, self.args.subject, self.args.message, from_=self.args.from_)


class LogLauncher(Launcher):

    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('log', help="Setup the server's database.")
        parser.add_argument('message', nargs='+', help='The log message')
        parser.add_argument('-g', '--logger', default='main', help='The logger name, default: main')
        parser.add_argument('-l', '--level', default='debug', help='Log level, default info')
        return parser

    def launch(self):
        from restfulpy.logging_ import get_logger

        getattr(get_logger(self.args.logger), self.args.level)(' '.join(self.args.message))


class ConfigurationDumpLauncher(Launcher):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('config-dump', help="Dump the configuration")
        parser.add_argument('path', help='The config path, for example: db')
        return parser

    def launch(self):
        dump = yaml.dump(getattr(settings, self.args.path), default_flow_style=False)
        print(dump)


class DevLauncher(Launcher, RequireSubCommand):
    @classmethod
    def create_parser(cls, subparsers):
        parser = subparsers.add_parser('dev', help="Development commands")
        dev_subparsers = parser.add_subparsers(title="admin command", dest="admin_command")
        LogLauncher.register(dev_subparsers)
        EmailLauncher.register(dev_subparsers)
        ConfigurationDumpLauncher.register(dev_subparsers)
        return parser




from sqlalchemy import Integer, ForeignKey, Unicode, JSON
from sqlalchemy.ext.declarative import declared_attr

from nanohttp import settings
from restfulpy.orm import Field
from restfulpy.taskqueue import Task
from restfulpy.logging_ import get_logger
from .providers import create_messenger

logger = get_logger('messaging')


# noinspection PyAbstractClass
class BaseEmail(Task):
    __abstract__ = True

    from_ = Field(Unicode(100), json='from', default=lambda: settings.messaging.default_sender)
    to = Field(Unicode(100), json='to')
    subject = Field(Unicode(256), json='subject')
    cc = Field(Unicode(100), nullable=True, json='cc')
    bcc = Field(Unicode(100), nullable=True, json='bcc')

    # noinspection PyDefaultArgument,PyMethodParameters
    @declared_attr
    def id(cls):
        return Field(Integer, ForeignKey('task.id'), primary_key=True, json='id')

    @property
    def email_body(self):  # pragma: no cover
        raise NotImplementedError()

    @property
    def template_filename(self):  # pragma: no cover
        raise NotImplementedError

    def do_(self, context, attachments=None):
        messenger = create_messenger()
        messenger.send(
            self.to,
            self.subject,
            self.email_body,
            cc=self.cc,
            bcc=self.bcc,
            template_filename=self.template_filename,
            from_=self.from_,
            attachments=attachments
        )

        logger.info('%s is sent to %s', self.subject, self.to)


# noinspection PyAbstractClass
class Email(BaseEmail):  # pragma: no cover
    __tablename__ = 'email'
    __mapper_args__ = {
        'polymorphic_identity': __tablename__
    }

    body = Field(JSON, json='body')

    @property
    def email_body(self):
        return self.body





import unittest

from nanohttp import json, RestController, context, settings
from sqlalchemy import Unicode, Integer

from restfulpy.controllers import JsonPatchControllerMixin
from restfulpy.orm import commit, DeclarativeBase, Field, DBSession
from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication


class CommitCheckingModel(DeclarativeBase):
    __tablename__ = 'commit_checking_model'
    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50), unique=True)


class Root(JsonPatchControllerMixin, RestController):

    @json
    @commit
    def post(self):
        m = CommitCheckingModel()
        m.title = context.form['title']
        DBSession.add(m)
        return m

    @json
    def get(self, title: str=None):
        m = DBSession.query(CommitCheckingModel).filter(CommitCheckingModel.title == title).one()
        return m

    @json
    @commit
    def error(self):
        m = CommitCheckingModel()
        m.title = 'Error'
        DBSession.add(m)
        raise Exception()


class CommitDecoratorTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', Root())
    __configuration__ = '''
    db:
      uri: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)

    def test_commit_decorator(self):
        self.request('ALL', 'POST', '/', params=dict(title='first'), doc=False)
        resp, ___ = self.request('ALL', 'GET', '/first', doc=False)
        self.assertEqual(resp['title'], 'first')
        self.assertEqual(resp['id'], 1)

    def test_commit_decorator_and_json_patch(self):
        # The commit decorator should not to do anything if the request is a jsonpatch.
        self.request('ALL', 'PATCH', '/', doc=False, json=[
            dict(op='post', path='', value=dict(title='second')),
            dict(op='post', path='', value=dict(title='third'))
        ])
        resp, ___ = self.request('ALL', 'GET', '/third', doc=False)
        self.assertEqual(resp['title'], 'third')

    def test_rollback(self):
        self.request('ALL', 'ERROR', '/', doc=False, expected_status=500)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




import io
import unittest
from os.path import dirname, abspath, join
from datetime import datetime, timezone, timedelta, time

from restfulpy.utils import import_python_module_by_filename, construct_class_by_name, format_iso_datetime, \
    format_iso_time, random_password, copy_stream, md5sum


HERE = abspath(dirname(__file__))
DATA_DIR = join(HERE, 'data')


class MyClassToConstructByName:
    def __init__(self, a):
        self.a = a


class UtilsTestCase(unittest.TestCase):

    def test_import_python_module_by_filename(self):
        filename = join(DATA_DIR, 'a.py')
        with open(filename, mode='w') as f:
            f.write('b = 123\n')

        module_ = import_python_module_by_filename('a', filename)
        self.assertEqual(module_.b, 123)

    def test_construct_class_by_name(self):
        obj = construct_class_by_name('restfulpy.tests.test_utils.MyClassToConstructByName', 1)
        self.assertEqual(obj.a, 1)
        self.assertIsNotNone(obj)

    def test_format_iso_datetime(self):
        result = format_iso_datetime(datetime(2012, 2, 22, 12, 52, 29, 300, tzinfo=timezone(timedelta(minutes=30))))
        self.assertEqual(result, '2012-02-22T12:52:29+00:30')

        result = format_iso_datetime(datetime(2012, 2, 22, 12, 52, 29, 300, tzinfo=timezone(timedelta(minutes=0))))
        self.assertEqual(result, '2012-02-22T12:52:29Z')

    def test_format_iso_time(self):
        result = format_iso_time(time(12, 52, 29, 300, tzinfo=timezone(timedelta(minutes=30))))
        self.assertEqual(result, '12:52:29+00:30')

        result = format_iso_time(time(12, 52, 29, 300, tzinfo=timezone(timedelta(minutes=0))))
        self.assertEqual(result, '12:52:29Z')

    def test_random_password(self):
        result = random_password(5)
        self.assertEqual(len(result), 5)

    def test_copy_stream(self):
        content = b'This is the initial source file'
        source = io.BytesIO(content)
        target = io.BytesIO()
        copy_stream(source, target)
        target.seek(0)
        self.assertEqual(target.read(), content)

    def test_md5sum(self):
        content = b'This is the initial source file'
        source = io.BytesIO(content)
        filename = join(DATA_DIR, 'a.txt')
        with open(filename, mode='wb') as f:
            f.write(content)

        self.assertEqual(md5sum(source), md5sum(filename))


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

from sqlalchemy import Integer, Unicode
from sqlalchemy.orm import synonym

from nanohttp import settings
from nanohttp.contexts import Context

from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication
from restfulpy.orm import DeclarativeBase, Field, DBSession, OrderingMixin, OrderableMixin


class OrderableOrderingObject(OrderableMixin, OrderingMixin, DeclarativeBase):
    __tablename__ = 'orderable_ordering_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))
    _age = Field(Integer)

    def _set_age(self, age):
        self._age = age

    def _get_age(self):  # pragma: no cover
        return self._age

    age = synonym('_age', descriptor=property(_get_age, _set_age))


class OrderingObject(OrderingMixin, DeclarativeBase):
    __tablename__ = 'ordering_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))


class OrderingMixinTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', None)
    __configuration__ = '''
    db:
      uri: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)

    def test_ordering_mixin(self):
        for i in range(1, 6):
            # noinspection PyArgumentList
            obj = OrderableOrderingObject(title='object %s' % i, age=i * 10)
            DBSession.add(obj)

            # noinspection PyArgumentList
            obj = OrderingObject(title='object %s' % i,)
            DBSession.add(obj)
        DBSession.commit()

        # Default soring with Orderable objects
        with Context({'QUERY_STRING': ''}, self.application):
            result = OrderableOrderingObject.sort_by_request().all()
            self.assertEqual(result[0].id, 1)
            self.assertEqual(result[-1].id, 5)

        # Default soring without Orderable objects
        with Context({'QUERY_STRING': ''}, self.application):
            result = OrderingObject.sort_by_request().all()
            self.assertEqual(len(result), 5)

        # Ascending
        with Context({'QUERY_STRING': 'sort=id'}, self.application):
            result = OrderableOrderingObject.sort_by_request().all()
            self.assertEqual(result[0].id, 1)
            self.assertEqual(result[-1].id, 5)

        # Descending
        with Context({'QUERY_STRING': 'sort=-id'}, self.application):
            result = OrderableOrderingObject.sort_by_request().all()
            self.assertEqual(result[0].id, 5)
            self.assertEqual(result[-1].id, 1)

        # Sort by Synonym Property
        with Context({'QUERY_STRING': 'sort=age'}, self.application):
            result = OrderableOrderingObject.sort_by_request().all()
            self.assertEqual(result[0].id, 1)
            self.assertEqual(result[-1].id, 5)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




import smtpd
import asyncore
import threading
import contextlib


SERVER_LOCK = threading.Event()


class MockupSMTPServer(smtpd.SMTPServer):

    def __init__(self, bind):
        super().__init__(bind, None, decode_data=False)
        self.server_address = self.socket.getsockname()[:2]
        SERVER_LOCK.set()

    def process_message(*args, **kwargs):
        pass


@contextlib.contextmanager
def smtp_server(bind=('localhost', 0)):
    server = MockupSMTPServer(bind)
    thread = threading.Thread(target=asyncore.loop, daemon=True)
    thread.start()
    SERVER_LOCK.wait()
    yield server, server.server_address
    # server.close()
    asyncore.close_all() #ignore_all=True)
    # thread.join()



import unittest
from datetime import date, time

from sqlalchemy import Integer, Unicode, ForeignKey, Boolean, Table, Date, Time, Float
from sqlalchemy.orm import synonym
from nanohttp import configure, HttpBadRequest

from restfulpy.orm import DeclarativeBase, init_model, create_engine, Field, DBSession, setup_schema, relationship, \
    composite, ModifiedMixin


class FullName(object):  # pragma: no cover
    def __init__(self, first_name, last_name):
        self.first_name = first_name
        self.last_name = last_name

    def __composite_values__(self):
        return '%s %s' % (self.first_name, self.last_name)

    def __repr__(self):
        return "FullName(%s %s)" % (self.first_name, self.last_name)

    def __eq__(self, other):
        return isinstance(other, FullName) and \
               other.first_name == self.first_name and \
               other.last_name == self.last_name

    def __ne__(self, other):
        return not self.__eq__(other)


class Author(DeclarativeBase):
    __tablename__ = 'author'

    id = Field(Integer, primary_key=True)
    email = Field(Unicode(100), unique=True, index=True, json='email',
                  pattern=r'(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)', watermark='Email',
                  example="user@example.com")
    title = Field(Unicode(50), index=True, min_length=2, watermark='First Name')
    first_name = Field(Unicode(50), index=True, json='firstName', min_length=2, watermark='First Name')
    last_name = Field(Unicode(100), json='lastName', min_length=2, watermark='Last Name')
    phone = Field(
        Unicode(10), nullable=True, json='phone', min_length=10,
        pattern=r'\d{3}[-\.\s]??\d{3}[-\.\s]??\d{4}|\(\d{3}\)\s*\d{3}[-\.\s]??\d{4}|\d{3}[-\.\s]??\d{4}',
        watermark='Phone'
    )
    name = composite(FullName, first_name, last_name, readonly=True, json='fullName', protected=True)
    _password = Field('password', Unicode(128), index=True, json='password', protected=True, min_length=6)
    birth = Field(Date)
    weight = Field(Float(asdecimal=True))

    def _set_password(self, password):
        self._password = 'hashed:%s' % password

    def _get_password(self):  # pragma: no cover
        return self._password

    password = synonym('_password', descriptor=property(_get_password, _set_password), info=dict(protected=True))


class Memo(DeclarativeBase):
    __tablename__ = 'memo'
    id = Field(Integer, primary_key=True)
    content = Field(Unicode(100), max_length=90)
    post_id = Field(ForeignKey('post.id'), json='postId')


class Comment(DeclarativeBase):
    __tablename__ = 'comment'
    id = Field(Integer, primary_key=True)
    content = Field(Unicode(100), max_length=90)
    special = Field(Boolean, default=True)
    post_id = Field(ForeignKey('post.id'), json='postId')
    post = relationship('Post')


post_tag_table = Table(
    'post_tag', DeclarativeBase.metadata,
    Field('post_id', Integer, ForeignKey('post.id')),
    Field('tag_id', Integer, ForeignKey('tag.id'))
)


class Tag(DeclarativeBase):
    __tablename__ = 'tag'
    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50), watermark='title', label='title')
    posts = relationship('Post', secondary=post_tag_table, back_populates='tags')


class Post(ModifiedMixin, DeclarativeBase):
    __tablename__ = 'post'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50), watermark='title', label='title', icon='star')
    author_id = Field(ForeignKey('author.id'), json='authorId')
    author = relationship(Author)
    memos = relationship(Memo, protected=True, json='privateMemos')
    comments = relationship(Comment)
    tags = relationship(Tag, secondary=post_tag_table, back_populates='posts')
    tag_time = Field(Time)


class ModelTestCase(unittest.TestCase):
    __configuration__ = '''
    db:
      uri: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def setUpClass(cls):
        configure(init_value=cls.__configuration__, force=True)
        cls.engine = create_engine()
        init_model(cls.engine)
        setup_schema()

    def test_model(self):
        # noinspection PyArgumentList
        author1 = Author(
            title='author1',
            email='author1@example.org',
            first_name='author 1 first name',
            last_name='author 1 last name',
            phone=None,
            password='123456',
            birth=date(1, 1, 1),
            weight=1.1
        )

        # noinspection PyArgumentList
        post1 = Post(
            title='First post',
            author=author1,
            tag_time=time(1, 1, 1)
        )
        DBSession.add(post1)
        DBSession.commit()

        self.assertEqual(post1.id, 1)

        # Validation, Type
        with self.assertRaises(HttpBadRequest):
            Author(title=234)

        # Validation, Pattern
        with self.assertRaises(HttpBadRequest):
            Author(email='invalidEmailAddress')

        # Validation, Min length
        with self.assertRaises(HttpBadRequest):
            Author(title='1')

        # Validation, Max length
        # Validation, Max length
        with self.assertRaises(HttpBadRequest):
            Author(phone='12321321321312321312312')

        # Metadata
        author_metadata = Author.json_metadata()
        self.assertIn('id', author_metadata)
        self.assertIn('email', author_metadata)
        self.assertNotIn('fullName', author_metadata)
        self.assertNotIn('password', author_metadata)

        post_metadata = Post.json_metadata()
        self.assertIn('author', post_metadata)

        comment_metadata = Comment.json_metadata()
        self.assertIn('post', comment_metadata)

        tag_metadata = Tag.json_metadata()
        self.assertIn('posts', tag_metadata)

        self.assertEqual(Comment.import_value(Comment.__table__.c.special, 'TRUE'), True)

        post1_dict = post1.to_dict()
        self.assertDictContainsSubset(
            {
                'author': {
                    'email': 'author1@example.org',
                    'firstName': 'author 1 first name',
                    'id': 1,
                    'lastName': 'author 1 last name',
                    'phone': None,
                    'title': 'author1',
                    'birth': '0001-01-01',
                    'weight': '1.1000000000'
                },
                'authorId': 1,
                'comments': [],
                'id': 1,
                'tags': [],
                'title': 'First post',
                'tagTime': '01:01:01'
            },
            post1_dict,
        )
        self.assertIn('createdAt', post1_dict)
        self.assertIn('modifiedAt', post1_dict)

        author1_dict = author1.to_dict()
        self.assertNotIn('fullName', author1_dict)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import copy
import unittest

from nanohttp import context, json, settings

from restfulpy.principal import DummyIdentity
from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication
from restfulpy.validation import validate_form
from restfulpy.controllers import RestController, RootController


class ValidationWhitelistController(RestController):

    @json
    @validate_form(whitelist=['whitelistParamForAll'],
                   client={'whitelist': ['whitelistParamForClient']},
                   admin={'whitelist': ['whitelistParamForAdmin']})
    def post(self):
        result = copy.deepcopy(context.form)
        result.update(context.query_string)
        return result


class Root(RootController):
    validation = ValidationWhitelistController()


class ValidationWhitelistTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', Root())

    @classmethod
    def configure_app(cls):
        super().configure_app()
        settings.merge("""
            logging:
              loggers:
                default:
                  level: info
            """)

    def test_validation_whitelist(self):
        # Test `whitelist`
        # role -> All
        self.wsgi_app.jwt_token = DummyIdentity().dump().decode()
        result, ___ = self.request('All', 'POST', '/validation', doc=False,
                                   params={'whitelistParamForAll': 'param'})
        self.assertIn('whitelistParamForAll', result)
        self.request('All', 'POST', '/validation', doc=False)
        self.request(
            'All', 'POST', '/validation', doc=False,
            params={
                'whitelistParamForAll': 'param',
                'customParam': 'param'
            },
            expected_status=400
        )
        self.request('All', 'POST', '/validation', doc=False, params={'customParam': 'param'},
                     expected_status=400)
        self.request('All', 'POST', '/validation', doc=False, params={'whitelistParamForClient': 'param', },
                     expected_status=400)
        # -----------------------------
        # role -> Client
        self.wsgi_app.jwt_token = DummyIdentity('client').dump().decode()

        result, ___ = self.request('Client', 'POST', '/validation', doc=False,
                                   params={'whitelistParamForAll': 'param'})
        self.assertIn('whitelistParamForAll', result)

        result, ___ = self.request('Client', 'POST', '/validation', doc=False,
                                   params={'whitelistParamForClient': 'param'})
        self.assertIn('whitelistParamForClient', result)

        result, ___ = self.request(
            'Client', 'POST', '/validation', doc=False,
            params={
                'whitelistParamForAll': 'param',
                'whitelistParamForClient': 'param',
            }
        )
        self.assertIn('whitelistParamForAll', result)
        self.assertIn('whitelistParamForClient', result)

        self.request('Client', 'POST', '/validation', doc=False)

        self.request(
            'Client', 'POST', '/validation', doc=False,
            params={
                'whitelistParamForAll': 'param',
                'customParam': 'param',
            },
            expected_status=400
        )

        self.request(
            'Client', 'POST', '/validation', doc=False,
            params={
                'customParam': 'param',
            },
            expected_status=400
        )

        self.request(
            'Client', 'POST', '/validation', doc=False,
            params={
                'whitelistParamForAdmin': 'param',
            },
            expected_status=400
        )

        # -----------------------------
        # role -> Admin
        self.wsgi_app.jwt_token = DummyIdentity('admin').dump().decode()

        result, ___ = self.request('Admin', 'POST', '/validation', doc=False,
                                   params={'whitelistParamForAll': 'param'})
        self.assertIn('whitelistParamForAll', result)

        result, ___ = self.request('Admin', 'POST', '/validation', doc=False,
                                   params={'whitelistParamForAdmin': 'param'})
        self.assertIn('whitelistParamForAdmin', result)

        result, ___ = self.request(
            'Admin', 'POST', '/validation', doc=False,
            params={
                'whitelistParamForAll': 'param',
                'whitelistParamForAdmin': 'param',
            }
        )
        self.assertIn('whitelistParamForAll', result)
        self.assertIn('whitelistParamForAdmin', result)

        self.request('Admin', 'POST', '/validation', doc=False)

        self.request(
            'Admin', 'POST', '/validation', doc=False,
            params={
                'whitelistParamForAll': 'param',
                'customParam': 'param',
            },
            expected_status=400
        )

        self.request('Admin', 'POST', '/validation', doc=False, params={
            'customParam': 'param',
        }, expected_status=400)

        self.request(
            'Admin', 'POST', '/validation', doc=False,
            params={
                'whitelistParamForAll': 'param',
                'whitelistParamForAdmin': 'param',
                'whitelistParamForClient': 'param',
            },
            expected_status=400
        )


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

from sqlalchemy import Integer, Unicode
from nanohttp import settings, HttpBadRequest
from nanohttp.contexts import Context

from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication
from restfulpy.orm import DeclarativeBase, Field, DBSession, FilteringMixin


class FilteringObject(FilteringMixin, DeclarativeBase):
    __tablename__ = 'filtering_object'

    id = Field(Integer, primary_key=True)
    title = Field(Unicode(50))


class FilteringMixinTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', None)
    __configuration__ = '''
    db:
      uri: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)

    def test_filtering_mixin(self):
        for i in range(1, 6):
            # noinspection PyArgumentList
            obj = FilteringObject(
                title='object %s' % i,
            )
            DBSession.add(obj)
        DBSession.commit()

        # Bad Value
        with Context({'QUERY_STRING': 'id=1'}, self.application) as context:
            context.query_string['id'] = 1
            self.assertRaises(HttpBadRequest, FilteringObject.filter_by_request)

        # IN
        with Context({'QUERY_STRING': 'id=^1,2,3'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 3)

        # NOT IN
        with Context({'QUERY_STRING': 'id=!^1,2,3'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 2)

        # IN (error)
        with Context({'QUERY_STRING': 'id=^'}, self.application):
            self.assertRaises(HttpBadRequest, FilteringObject.filter_by_request)

        # Between
        with Context({'QUERY_STRING': 'id=~1,3'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 3)

        # IS NULL
        with Context({'QUERY_STRING': 'title=null'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 0)

        # IS NOT NULL
        with Context({'QUERY_STRING': 'title=!null'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 5)

        # LIKE
        with Context({'QUERY_STRING': 'title=%obj'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 5)

        with Context({'QUERY_STRING': 'title=%OBJ'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 0)

        # ILIKE
        with Context({'QUERY_STRING': 'title=%~obj'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 5)

        with Context({'QUERY_STRING': 'title=%~OBJ'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 5)

        # ==
        with Context({'QUERY_STRING': 'id=1'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 1)

        # !=
        with Context({'QUERY_STRING': 'id=!1'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 4)

        # >=
        with Context({'QUERY_STRING': 'id=>=2'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 4)

        # >
        with Context({'QUERY_STRING': 'id=>2'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 3)

        # <=
        with Context({'QUERY_STRING': 'id=<=3'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 3)

        # <
        with Context({'QUERY_STRING': 'id=<3'}, self.application):
            self.assertEqual(FilteringObject.filter_by_request().count(), 2)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()




import unittest
import threading

from restfulpy.taskqueue import Task, worker
from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication


awesome_task_done = threading.Event()
another_task_done = threading.Event()


class AwesomeTask(Task):

    __mapper_args__ = {
        'polymorphic_identity': 'awesome_task'
    }

    def do_(self, context):
        awesome_task_done.set()


class AnotherTask(Task):

    __mapper_args__ = {
        'polymorphic_identity': 'another_task'
    }

    def do_(self, context):
        another_task_done.set()


class BadTask(Task):

    __mapper_args__ = {
        'polymorphic_identity': 'bad_task'
    }

    def do_(self, context):
        raise Exception()


class TaskQueueTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', None)

    def test_worker(self):
        # noinspection PyArgumentList
        awesome_task = AwesomeTask()
        self.session.add(awesome_task)

        another_task = AnotherTask()
        self.session.add(another_task)

        bad_task = BadTask()
        self.session.add(bad_task)

        self.session.commit()

        tasks = worker(tries=0, filters=Task.type == 'awesome_task')
        self.assertEqual(len(tasks), 1)
        self.assertTrue(awesome_task_done.is_set())
        self.assertFalse(another_task_done.is_set())

        self.session.refresh(awesome_task)
        self.assertEqual(awesome_task.status, 'success')

        tasks = worker(tries=0, filters=Task.type == 'bad_task')
        self.assertEqual(len(tasks), 1)
        bad_task_id = tasks[0][0]
        self.session.refresh(bad_task)
        self.assertEqual(bad_task.status, 'failed')

        tasks = worker(tries=0, filters=Task.type == 'bad_task')
        self.assertEqual(len(tasks), 0)

        # Reset the status of one task
        self.session.refresh(bad_task)
        bad_task.status = 'in-progress'
        self.session.commit()
        self.session.refresh(bad_task)

        Task.reset_status(bad_task_id, self.session)
        self.session.commit()
        tasks = worker(tries=0, filters=Task.type == 'bad_task')
        self.assertEqual(len(tasks), 1)

        tasks = worker(tries=0, filters=Task.type == 'bad_task')
        self.assertEqual(len(tasks), 0)

        # Cleanup all tasks
        Task.cleanup(self.session, statuses=('in-progress', 'failed'))
        self.session.commit()

        tasks = worker(tries=0, filters=Task.type == 'bad_task')
        self.assertEqual(len(tasks), 1)

        tasks = worker(tries=0, filters=Task.type == 'bad_task')
        self.assertEqual(len(tasks), 0)

        # Doing all remaining tasks
        tasks = worker(tries=0)
        self.assertEqual(len(tasks), 1)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import time
import traceback
from datetime import datetime

from sqlalchemy import Integer, Enum, Unicode, DateTime
from sqlalchemy.sql.expression import text
from nanohttp import settings

from restfulpy.orm import TimestampMixin, DeclarativeBase, Field, DBSession, create_thread_unsafe_session
from restfulpy.exceptions import RestfulException
from restfulpy.logging_ import get_logger

logger = get_logger('taskqueue')


class TaskPopError(RestfulException):
    pass


class Task(TimestampMixin, DeclarativeBase):
    __tablename__ = 'task'

    id = Field(Integer, primary_key=True, json='id')
    priority = Field(Integer, nullable=False, default=50, json='priority')
    status = Field(Enum('new', 'success', 'in-progress', 'failed', name='task_status_enum'), default='new',
                   nullable=True, json='status')
    fail_reason = Field(Unicode(2048), nullable=True, json='reason')
    started_at = Field(DateTime, nullable=True, json='startedAt')
    terminated_at = Field(DateTime, nullable=True, json='terminatedAt')
    type = Field(Unicode(50))

    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
        'polymorphic_on': type
    }

    def do_(self, context):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def pop(cls, statuses={'new'}, filters=None, session=DBSession):

        find_query = session.query(cls.id.label('id'), cls.created_at, cls.status, cls.type, cls.priority)
        if filters is not None:
            find_query = find_query.filter(text(filters) if isinstance(filters, str) else filters)

        find_query = find_query \
            .filter(cls.status.in_(statuses)) \
            .order_by(cls.priority.desc()) \
            .order_by(cls.created_at) \
            .limit(1) \
            .with_lockmode('update')

        cte = find_query.cte('find_query')

        update_query = Task.__table__.update() \
            .where(Task.id == cte.c.id) \
            .values(status='in-progress') \
            .returning(Task.__table__.c.id)

        task_id = session.execute(update_query).fetchone()
        session.commit()
        if not task_id:
            raise TaskPopError('There is no task to pop')
        task_id = task_id[0]
        task = session.query(cls).filter(cls.id == task_id).one()
        return task

    def execute(self, context, session=DBSession):
        try:
            isolated_task = session.query(Task).filter(Task.id == self.id).one()
            isolated_task.do_(context)
            session.commit()
        except:
            session.rollback()
            raise

    @classmethod
    def cleanup(cls, session=DBSession, statuses=['in-progress']):
        session.query(Task) \
            .filter(Task.status.in_(statuses)) \
            .with_lockmode('update') \
            .update({'status': 'new', 'started_at': None, 'terminated_at': None}, synchronize_session='fetch')

    @classmethod
    def reset_status(cls, task_id, session=DBSession, statuses=['in-progress']):
        session.query(Task) \
            .filter(Task.status.in_(statuses)) \
            .filter(Task.id == task_id) \
            .with_lockmode('update') \
            .update({'status': 'new', 'started_at': None, 'terminated_at': None}, synchronize_session='fetch')


def worker(statuses={'new'}, filters=None, tries=-1):
    isolated_session = create_thread_unsafe_session()
    context = {'counter': 0}
    tasks = []
    while True:
        context['counter'] += 1
        logger.info("Trying to pop a task, Counter: %s" % context['counter'])

        try:
            task = Task.pop(
                statuses=statuses,
                filters=filters,
                session=isolated_session
            )

        except TaskPopError as ex:
            logger.info('No task to pop: %s' % ex.to_json())
            isolated_session.rollback()
            if tries > -1:
                tries -= 1
                if tries <= 0:
                    return tasks

        # noinspection PyBroadException
        try:
            task.execute(context)

            # Task success
            task.status = 'success'
            task.terminated_at = datetime.now()

        except:
            logger.exception('Error when executing task: %s' % task.id)
            task.status = 'failed'
            task.fail_reason = traceback.format_exc()

        finally:
            if isolated_session.is_active:
                isolated_session.commit()
            tasks.append((task.id, task.status))

        time.sleep(settings.worker.gap)




import unittest

from nanohttp import configure

from restfulpy.principal import JwtPrincipal


class PrincipalTestCase(unittest.TestCase):
    __configuration__ = '''
    jwt:
      secret: JWT-SECRET
      algorithm: HS256
      max_age: 86400  # 24 Hours
      refresh_token:
        secret: JWT-REFRESH-SECRET
        algorithm: HS256
        max_age: 2678400  # 30 Days
    '''

    @classmethod
    def setUpClass(cls):
        configure(init_value=cls.__configuration__, force=True)

    def test_principal(self):
        principal = JwtPrincipal(dict(
            email='test@example.com',
            id=1,
            sessionId=1,
            roles=['admin']
        ))

        self.assertEqual(principal.email, 'test@example.com')
        self.assertEqual(principal.id, 1)
        self.assertEqual(principal.session_id, 1)
        self.assertEqual(principal.roles, ['admin'])
        self.assertTrue(principal.is_in_roles('admin'))
        self.assertTrue(principal.is_in_roles('admin', 'god'))

        encoded = principal.dump()

        principal = JwtPrincipal.load(encoded.decode())
        self.assertEqual(principal.email, 'test@example.com')
        self.assertEqual(principal.id, 1)
        self.assertEqual(principal.session_id, 1)
        self.assertEqual(principal.roles, ['admin'])
        self.assertTrue(principal.is_in_roles('admin'))
        self.assertTrue(principal.is_in_roles('admin', 'god'))

        principal = JwtPrincipal.load(encoded.decode(), force=True)
        self.assertEqual(principal.email, 'test@example.com')
        self.assertEqual(principal.id, 1)
        self.assertEqual(principal.session_id, 1)
        self.assertEqual(principal.roles, ['admin'])
        self.assertTrue(principal.is_in_roles('admin'))
        self.assertTrue(principal.is_in_roles('admin', 'god'))

        principal = JwtPrincipal.load((b'Bearer %s' % encoded).decode(), force=True)
        self.assertEqual(principal.email, 'test@example.com')
        self.assertEqual(principal.id, 1)
        self.assertEqual(principal.session_id, 1)
        self.assertEqual(principal.roles, ['admin'])
        self.assertTrue(principal.is_in_roles('admin'))
        self.assertTrue(principal.is_in_roles('admin', 'god'))


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

from sqlalchemy import Unicode
from nanohttp import settings, HttpConflict

from restfulpy.orm import DeclarativeBase, DBSession, Field, SoftDeleteMixin
from restfulpy.testing import WebAppTestCase
from restfulpy.tests.helpers import MockupApplication


class SoftDeleteCheckingModel(SoftDeleteMixin, DeclarativeBase):
    __tablename__ = 'soft_delete_checking_model'

    title = Field(Unicode(50), primary_key=True)


class SoftDeleteCheckingModelTestCase(WebAppTestCase):
    application = MockupApplication('MockupApplication', None)
    __configuration__ = '''
    db:
      uri: sqlite://    # In memory DB
      echo: false
    '''

    @classmethod
    def configure_app(cls):
        cls.application.configure(force=True)
        settings.merge(cls.__configuration__)

    def test_soft_delete_mixin(self):
        # noinspection PyArgumentList
        instance = SoftDeleteCheckingModel(
            title='test title'
        )
        DBSession.add(instance)
        DBSession.commit()
        instance.assert_is_not_deleted()
        self.assertRaises(ValueError, instance.assert_is_deleted)

        instance = SoftDeleteCheckingModel.query.one()
        instance.soft_delete()
        DBSession.commit()
        instance.assert_is_deleted()
        self.assertRaises(ValueError, instance.assert_is_not_deleted)

        self.assertEqual(SoftDeleteCheckingModel.filter_deleted().count(), 1)
        self.assertEqual(SoftDeleteCheckingModel.exclude_deleted().count(), 0)

        instance.soft_undelete()
        DBSession.commit()
        instance.assert_is_not_deleted()
        self.assertRaises(ValueError, instance.assert_is_deleted)

        self.assertEqual(SoftDeleteCheckingModel.filter_deleted().count(), 0)
        self.assertEqual(SoftDeleteCheckingModel.exclude_deleted().count(), 1)

        DBSession.delete(instance)
        self.assertRaises(HttpConflict, DBSession.commit)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()
